"""This ingredient provides (expert) demonstrations to learn from.

The demonstrations are either loaded from disk, from the HuggingFace Dataset Hub, or
sampled from the expert policy provided by the expert ingredient.
"""

import logging
from typing import Any, Dict, Optional, Sequence

import datasets
import huggingface_sb3 as hfsb3
import numpy as np
import sacred

from imitation.data import huggingface_utils, rollout, serialize, types
from imitation.scripts.ingredients import environment, expert
from imitation.scripts.ingredients import logging as logging_ingredient

demonstrations_ingredient = sacred.Ingredient(
    "demonstrations",
    ingredients=[
        expert.expert_ingredient,
        logging_ingredient.logging_ingredient,
        environment.environment_ingredient,
    ],
)
logger = logging.getLogger(__name__)


@demonstrations_ingredient.config
def config():
    # Either "local" or "huggingface" or "generated".
    source = "generated"

    # local path or huggingface repo id to load rollouts from.
    path = None

    # passed to `datasets.load_dataset` if `source` is "huggingface"
    loader_kwargs: Dict[str, Any] = dict(
        split="train",
    )

    # Used to deduce HuggingFace repo id if `path` is None
    organization = "HumanCompatibleAI"

    # Used to deduce HuggingFace repo id if `path` is None
    algo_name = "ppo"

    # Num demos used or sampled. None loads every demo possible.
    n_expert_demos = None
    locals()  # quieten flake8


@demonstrations_ingredient.named_config
def fast():
    # Note: we can't pick `n_expert_demos=1` here because for envs with short episodes
    #   that does not generate the minimum number of transitions required for one batch.
    n_expert_demos = 10  # noqa: F841


@demonstrations_ingredient.capture
def get_expert_trajectories(
    source: str,
    path: str,
) -> Sequence[types.Trajectory]:
    """Loads expert demonstrations.

    Args:
        source: Can be either `local` to load rollouts from the disk,
            `huggingface` to load from the HuggingFace hub or
            `generated` to generate the expert trajectories.
        path: A path containing a pickled sequence of `sources.Trajectory`.

    Returns:
        The expert trajectories.

    Raises:
        ValueError: if `source` is not in ["local", "huggingface", "generated"].
    """
    if source == "local":
        if path is None:
            raise ValueError(
                "When source is 'local', path must be set.",
            )
        return _constrain_number_of_demos(serialize.load(path))

    if source == "huggingface":
        return _constrain_number_of_demos(_download_expert_rollouts())

    if source == "generated":
        if path is not None:
            logger.warning("Ignoring path when source is 'generated'")
        return _generate_expert_trajs()

    raise ValueError(
        "`source` can either be `local` or `huggingface` or `generated`.",
    )


@demonstrations_ingredient.capture
def _constrain_number_of_demos(
    demos: Sequence[types.Trajectory],
    n_expert_demos: Optional[int],
) -> Sequence[types.Trajectory]:
    """Constrains the number of demonstrations to n_expert_demos if it is not None."""
    if n_expert_demos is None:
        return demos
    else:
        if len(demos) < n_expert_demos:
            raise ValueError(
                f"Want to use n_expert_demos={n_expert_demos} trajectories, but only "
                f"{len(demos)} are available.",
            )
        if len(demos) > n_expert_demos:
            logger.warning(
                f"Using only the first {n_expert_demos} trajectories out of "
                f"{len(demos)} available.",
            )
            return demos[:n_expert_demos]
        else:
            return demos


@demonstrations_ingredient.capture
def _generate_expert_trajs(
    n_expert_demos: Optional[int],
    _rnd: np.random.Generator,
) -> Optional[Sequence[types.Trajectory]]:
    """Generates expert demonstrations.

    Args:
        n_expert_demos: The number of trajectories to load.
            Dataset is truncated to this length if specified.
        _rnd: Random number generator provided by Sacred.

    Returns:
        The expert trajectories.

    Raises:
        ValueError: If n_expert_demos is None.
    """
    if n_expert_demos is None:
        raise ValueError("n_expert_demos must be specified when generating demos.")

    logger.info(f"Generating {n_expert_demos} expert trajectories")
    with environment.make_rollout_venv() as env:  # type: ignore[wrong-arg-count]
        return rollout.rollout(
            expert.get_expert_policy(env),
            env,
            rollout.make_sample_until(min_episodes=n_expert_demos),
            rng=_rnd,
        )


@demonstrations_ingredient.capture
def _download_expert_rollouts(
    environment: Dict[str, Any],
    path: Optional[str],
    organization: Optional[str],
    algo_name: Optional[str],
    loader_kwargs: Dict[str, Any],
):
    if path is not None:
        repo_id = path
    else:
        model_name = hfsb3.ModelName(
            algo_name,
            hfsb3.EnvironmentName(environment["gym_id"]),
        )
        repo_id = hfsb3.ModelRepoId(organization, model_name)

    logger.info(f"Loading expert trajectories from {repo_id}")
    dataset = datasets.load_dataset(repo_id, **loader_kwargs)
    return huggingface_utils.TrajectoryDatasetSequence(dataset)



"""Tests for imitation.util.sacred_file_parsing."""
import json
import pathlib

import imitation.util.sacred_file_parsing as sfp


def _make_sacred_run_dir(
    path: pathlib.Path,
    algo: str,
    env: str,
    status: str = "COMPLETED",
):
    path.mkdir(parents=True, exist_ok=True)
    cfg_file = path / "config.json"
    cfg_file.write_text(json.dumps(dict(environment=dict(gym_id=env))))

    run_file = path / "run.json"
    run_file.write_text(json.dumps(dict(status=status, command=algo)))


def test_load_single_run(tmp_path):
    # GIVEN
    _make_sacred_run_dir(tmp_path / "run1", "ppo", "CartPole-v1")

    # WHEN
    runs = list(sfp.find_sacred_runs(tmp_path))

    # THEN
    assert len(runs) == 1
    assert runs[0][0]["environment"]["gym_id"] == "CartPole-v1"
    assert runs[0][1]["command"] == "ppo"


def test_load_multiple_runs_in_sub_folders(tmp_path):
    # GIVEN
    _make_sacred_run_dir(tmp_path / "run1", "ppo", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "subfolder1" / "run2", "ppo", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "subfolder1" / "run3", "ppo", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "subfolder2" / "run4", "ppo", "CartPole-v1")

    # WHEN
    runs = list(sfp.find_sacred_runs(tmp_path))

    # THEN
    assert len(runs) == 4
    for conf, run in runs:
        assert conf["environment"]["gym_id"] == "CartPole-v1"
        assert run["command"] == "ppo"


def test_loading_only_completed_runs(tmp_path):
    # GIVEN
    _make_sacred_run_dir(tmp_path / "run1", "ppo", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "run2", "airl", "CartPole-v1", status="FAILED")
    _make_sacred_run_dir(tmp_path / "run3", "ppo", "CartPole-v1", status="COMPLETED")
    _make_sacred_run_dir(tmp_path / "run4", "gail", "CartPole-v1", status="RUNNING")

    # WHEN
    runs = list(sfp.find_sacred_runs(tmp_path, only_completed_runs=True))

    # THEN
    assert len(runs) == 2
    for conf, run in runs:
        assert conf["environment"]["gym_id"] == "CartPole-v1"
        assert run["command"] == "ppo"


def test_grouping_runs_by_algo_and_env(tmp_path):
    # GIVEN
    _make_sacred_run_dir(tmp_path / "run1", "ppo", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "run2", "airl", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "run3", "ppo", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "run4", "gail", "CartPole-v1")
    _make_sacred_run_dir(tmp_path / "run5", "ppo", "LunarLander-v2")
    _make_sacred_run_dir(tmp_path / "run6", "airl", "LunarLander-v2")
    _make_sacred_run_dir(tmp_path / "run7", "ppo", "LunarLander-v2")

    # WHEN
    runs_by_algo_and_env = sfp.group_runs_by_algo_and_env(tmp_path)

    # THEN
    assert set(runs_by_algo_and_env.keys()) == {"ppo", "airl", "gail"}
    assert set(runs_by_algo_and_env["ppo"].keys()) == {"CartPole-v1", "LunarLander-v2"}
    assert set(runs_by_algo_and_env["airl"].keys()) == {"CartPole-v1", "LunarLander-v2"}
    assert set(runs_by_algo_and_env["gail"].keys()) == {"CartPole-v1"}



"""Types and helper methods for transitions and trajectories."""

import collections
import dataclasses
import itertools
import numbers
import os
import warnings
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
    overload,
)

import numpy as np
import torch as th
from torch.utils import data as th_data

T = TypeVar("T")

AnyPath = Union[str, bytes, os.PathLike]
AnyTensor = Union[np.ndarray, th.Tensor]
TensorVar = TypeVar("TensorVar", np.ndarray, th.Tensor)


@dataclasses.dataclass(frozen=True)
class DictObs:
    """Stores observations from an environment with a dictionary observation space.

    Provides an interface that is similar to observations in a numpy array.
    Length, slicing, indexing, and iterating operations will operate on the first
    dimension of the constituent arrays, as they would for observations in a single
    array.

    There are also utility functions for mapping / stacking / concatenating
    lists of dictobs.
    """

    _d: Dict[str, np.ndarray]

    @classmethod
    def from_obs_list(cls, obs_list: List[Dict[str, np.ndarray]]) -> "DictObs":
        """Stacks the observation list into a single DictObs."""
        return cls.stack(map(cls, obs_list))

    def __post_init__(self):
        if not all(
            isinstance(v, (np.ndarray, numbers.Number)) for v in self._d.values()
        ):
            raise TypeError("Values must be NumPy arrays")

    def __len__(self):
        """Returns the first dimension of constituent arrays.

        Only defined if there is at least one array, and all arrays have the same
        length of first dimension. Otherwise raises ValueError.

        Len of a DictObs usually represents number of timesteps, or number of
        environments in a VecEnv.

        Use `dict_len` to get the number of entries in the dictionary.

        Raises:
            RuntimeError: if the arrays have different lengths or there are no arrays.

        Returns:
            The length (first dimension) of the constituent arrays
        """
        lens = set(len(v) for v in self._d.values())
        if len(lens) == 1:
            return lens.pop()
        elif len(lens) == 0:
            raise RuntimeError("Length not defined as DictObs is empty")
        else:
            raise RuntimeError(
                f"Length not defined; arrays have conflicting first dimensions: {lens}",
            )

    @property
    def dict_len(self):
        """Returns the number of arrays in the DictObs."""
        return len(self._d)

    def __getitem__(
        self,
        key: Union[int, slice, Tuple[Union[int, slice], ...]],
    ) -> "DictObs":
        """Indexes or slices each array.

        See `.get` for accessing a value from the underlying dictionary.

        Note that it will still return singleton values as np.arrays, not scalars,
        to be consistent with DictObs type signature.

        Args:
            key: a single slice

        Returns:
            A new DictObj object with each array indexed.
        """
        # asarray handles case where we slice to a single array element.
        return self.__class__({k: np.asarray(v[key]) for k, v in self._d.items()})

    def __iter__(self) -> Iterator["DictObs"]:
        """Iterates over the first dimension of each array.

        Raises:
            ValueError if len() is not defined.

        Returns:
            Iterator of dictobjs by first dimension.
        """
        return (self[i] for i in range(len(self)))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not self.keys() == other.keys():
            return False
        return all(np.array_equal(self.get(k), other.get(k)) for k in self.keys())

    @property
    def shape(self) -> Dict[str, Tuple[int, ...]]:
        """Returns a dictionary with shape-tuples in place of the arrays."""
        return {k: v.shape for k, v in self.items()}

    @property
    def dtype(self) -> Dict[str, np.dtype]:
        """Returns a dictionary with dtype-tuples in place of the arrays."""
        return {k: v.dtype for k, v in self.items()}

    def keys(self):
        return self._d.keys()

    def values(self):
        return self._d.values()

    def items(self):
        return self._d.items()

    def __contains__(self, key):
        return key in self._d

    def get(self, key: str) -> np.ndarray:
        """Returns the array for the given key, or raises KeyError."""
        return self._d[key]

    def unwrap(self) -> Dict[str, np.ndarray]:
        """Returns a copy of the underlying dictionary (arrays are not copied)."""
        return dict(self._d)

    def map_arrays(self, fn: Callable[[np.ndarray], np.ndarray]) -> "DictObs":
        """Returns a new DictObs with `fn` applied to every array."""
        return self.__class__({k: fn(v) for k, v in self.items()})

    @staticmethod
    def _unravel(dictobs_list: Iterable["DictObs"]) -> Dict[str, List[np.ndarray]]:
        """Converts a list of DictObs into a dictionary of lists of arrays."""
        it1, it2 = itertools.tee(dictobs_list)
        # assert all have same keys
        key_set = set(frozenset(obs.keys()) for obs in it1)
        if len(key_set) == 0:
            raise ValueError("Empty list of DictObs")
        if not len(key_set) == 1:
            raise ValueError(f"Inconsistent keys: {key_set}")

        unraveled: Dict[str, List[np.ndarray]] = collections.defaultdict(list)
        for ob_dict in it2:
            for k, array in ob_dict._d.items():
                unraveled[k].append(array)
        return unraveled

    @classmethod
    def stack(cls, dictobs_list: Iterable["DictObs"], axis=0) -> "DictObs":
        """Returns a single dictobs stacking the arrays by key."""
        return cls(
            {
                k: np.stack(arr_list, axis=axis)
                for k, arr_list in cls._unravel(dictobs_list).items()
            },
        )

    @classmethod
    def concatenate(cls, dictobs_list: Iterable["DictObs"], axis=0) -> "DictObs":
        """Returns a single dictobs concatenating the arrays by key."""
        return cls(
            {
                k: np.concatenate(arr_list, axis=axis)
                for k, arr_list in cls._unravel(dictobs_list).items()
            },
        )


# DictObs utilities


Observation = Union[np.ndarray, DictObs]
ObsVar = TypeVar("ObsVar", np.ndarray, DictObs)


def assert_not_dictobs(x: Observation) -> np.ndarray:
    """Typeguard to assert `x` is an array, not a DictObs."""
    assert not isinstance(x, DictObs), "Dictionary observations are not supported here."
    return x


def concatenate_maybe_dictobs(arrs: List[ObsVar]) -> ObsVar:
    """Concatenates a list of observations appropriately (depending on type)."""
    assert len(arrs) > 0
    if isinstance(arrs[0], DictObs):
        return DictObs.concatenate(arrs)
    else:
        return np.concatenate(arrs)


def stack_maybe_dictobs(arrs: List[ObsVar]) -> ObsVar:
    """Stacks a list of observations appropriately (depending on type)."""
    assert len(arrs) > 0
    if isinstance(arrs[0], DictObs):
        return DictObs.stack(arrs)
    else:
        return np.stack(arrs)


# the following overloads have a type error as a DictObs matches both definitions, but
# the return types are incompatible. Ideally T would exclude DictObs but that's not
# possible.
@overload
def maybe_unwrap_dictobs(  # type: ignore[misc]
    maybe_dictobs: DictObs,
) -> Dict[str, np.ndarray]:
    ...


@overload
def maybe_unwrap_dictobs(maybe_dictobs: T) -> T:
    ...


def maybe_unwrap_dictobs(maybe_dictobs):
    """Unwraps if a DictObs, otherwise returns the object."""
    if isinstance(maybe_dictobs, DictObs):
        return maybe_dictobs.unwrap()
    else:
        if not isinstance(maybe_dictobs, (np.ndarray, th.Tensor, int)):
            warnings.warn(f"trying to unwrap object of type {type(maybe_dictobs)}")
        return maybe_dictobs


@overload
def maybe_wrap_in_dictobs(obs: Union[Dict[str, np.ndarray], DictObs]) -> DictObs:
    ...


@overload
def maybe_wrap_in_dictobs(obs: np.ndarray) -> np.ndarray:
    ...


def maybe_wrap_in_dictobs(
    obs: Union[Dict[str, np.ndarray], np.ndarray, DictObs],
) -> Observation:
    """Converts an observation into a DictObs, if necessary."""
    if isinstance(obs, dict):
        return DictObs(obs)
    else:
        if not isinstance(obs, (np.ndarray, DictObs, float, int)):
            warnings.warn(f"tried to wrap {type(obs)} as an observation")
        return obs


def map_maybe_dict(fn, maybe_dict):
    """Either maps fn over dictionary values or applies fn to `maybe_dict`.

    Args:
        fn: function to apply. Must take a single argument.
        maybe_dict: either a dict or a value that can be passed to fn.

    Returns:
        Either a dict (if maybe_dict was a dict) or `fn(maybe_dict)`.
    """
    if isinstance(maybe_dict, dict):
        return {k: fn(v) for k, v in maybe_dict.items()}
    else:
        return fn(maybe_dict)


class TransitionMappingNoNextObs(TypedDict):
    """Dictionary with `obs` and `acts`."""

    obs: Union[Observation, th.Tensor]
    acts: AnyTensor


# inheritance with total=False so these are not required
class TransitionMapping(TransitionMappingNoNextObs, total=False):
    """Dictionary with `obs` and `acts`, maybe also `next_obs`, `dones`, `rew`."""

    next_obs: Union[Observation, th.Tensor]
    dones: AnyTensor
    rew: AnyTensor


def dataclass_quick_asdict(obj) -> Dict[str, Any]:
    """Extract dataclass to items using `dataclasses.fields` + dict comprehension.

    This is a quick alternative to `dataclasses.asdict`, which expensively and
    undocumentedly deep-copies every numpy array value.
    See https://stackoverflow.com/a/52229565/1091722.

    This is also used to preserve DictObj objects, as `dataclasses.asdict`
    unwraps them recursively.

    Args:
        obj: A dataclass instance.

    Returns:
        A dictionary mapping from `obj` field names to values.
    """
    d = {f.name: getattr(obj, f.name) for f in dataclasses.fields(obj)}
    return d


@dataclasses.dataclass(frozen=True)
class Trajectory:
    """A trajectory, e.g. a one episode rollout from an expert policy."""

    obs: Observation
    """Observations, shape (trajectory_len + 1, ) + observation_shape."""

    acts: np.ndarray
    """Actions, shape (trajectory_len, ) + action_shape."""

    infos: Optional[np.ndarray]
    """An array of info dicts, shape (trajectory_len, ).

    The info dict is returned by some environments `step()` and contains auxiliary
    diagnostic information. For example the monitor wrapper adds an info dict
    to the last step of each episode containing the episode return and length.
    """

    terminal: bool
    """Does this trajectory (fragment) end in a terminal state?

    Episodes are always terminal. Trajectory fragments are also terminal when they
    contain the final state of an episode (even if missing the start of the episode).
    """

    def __len__(self) -> int:
        """Returns number of transitions, equal to the number of actions."""
        return len(self.acts)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Trajectory):
            return False

        dict_self = dataclass_quick_asdict(self)
        dict_other = dataclass_quick_asdict(other)
        # Trajectory objects may still have different keys if different subclasses
        if dict_self.keys() != dict_other.keys():
            return False

        if len(self) != len(other):
            # Short-circuit: if trajectories are of different length, then unequal.
            # Redundant as later checks would catch this, but speeds up common case.
            return False

        for k, self_v in dict_self.items():
            other_v = dict_other[k]
            if k == "infos":
                # Treat None equivalent to sequence of empty dicts
                self_v = [{}] * len(self) if self_v is None else self_v
                other_v = [{}] * len(other) if other_v is None else other_v
            if isinstance(self_v, DictObs):
                if not self_v == other_v:
                    return False
            if not np.array_equal(self_v, other_v):
                return False

        return True

    def __post_init__(self):
        """Performs input validation: check shapes are as specified in docstring."""
        if len(self.obs) != len(self.acts) + 1:
            raise ValueError(
                "expected one more observations than actions: "
                f"{len(self.obs)} != {len(self.acts)} + 1",
            )
        if self.infos is not None and len(self.infos) != len(self.acts):
            raise ValueError(
                "infos when present must be present for each action: "
                f"{len(self.infos)} != {len(self.acts)}",
            )
        if len(self.acts) == 0:
            raise ValueError("Degenerate trajectory: must have at least one action.")

    def __setstate__(self, state):
        if "terminal" not in state:
            warnings.warn(
                "Loading old version of Trajectory."
                "Support for this will be removed in future versions.",
                DeprecationWarning,
            )
            state["terminal"] = True
        self.__dict__.update(state)


def _rews_validation(rews: np.ndarray, acts: np.ndarray):
    if rews.shape != (len(acts),):
        raise ValueError(
            "rewards must be 1D array, one entry for each action: "
            f"{rews.shape} != ({len(acts)},)",
        )
    if not np.issubdtype(rews.dtype, np.floating):
        raise ValueError(f"rewards dtype {rews.dtype} not a float")


@dataclasses.dataclass(frozen=True, eq=False)
class TrajectoryWithRew(Trajectory):
    """A `Trajectory` that additionally includes reward information."""

    rews: np.ndarray
    """Reward, shape (trajectory_len, ). dtype float."""

    def __post_init__(self):
        """Performs input validation, including for rews."""
        super().__post_init__()
        _rews_validation(self.rews, self.acts)


Pair = Tuple[T, T]
TrajectoryPair = Pair[Trajectory]
TrajectoryWithRewPair = Pair[TrajectoryWithRew]


def transitions_collate_fn(
    batch: Sequence[Mapping[str, np.ndarray]],
) -> Mapping[str, AnyTensor]:
    """Custom `torch.utils.data.DataLoader` collate_fn for `TransitionsMinimal`.

    Use this as the `collate_fn` argument to `DataLoader` if using an instance of
    `TransitionsMinimal` as the `dataset` argument.

    Args:
        batch: The batch to collate.

    Returns:
        A collated batch. Uses Torch's default collate function for everything
        except the "infos" key. For "infos", we join all the info dicts into a
        list of dicts. (The default behavior would recursively collate every
        info dict into a single dict, which is incorrect.)
    """
    batch_acts_and_dones = [
        {k: np.array(v) for k, v in sample.items() if k in ["acts", "dones"]}
        for sample in batch
    ]

    result = th_data.dataloader.default_collate(batch_acts_and_dones)
    assert isinstance(result, dict)
    result["infos"] = [sample["infos"] for sample in batch]
    result["obs"] = stack_maybe_dictobs([sample["obs"] for sample in batch])
    result["next_obs"] = stack_maybe_dictobs([sample["next_obs"] for sample in batch])
    return result


TransitionsMinimalSelf = TypeVar("TransitionsMinimalSelf", bound="TransitionsMinimal")


@dataclasses.dataclass(frozen=True)
class TransitionsMinimal(th_data.Dataset, Sequence[Mapping[str, np.ndarray]]):
    """A Torch-compatible `Dataset` of obs-act transitions.

    This class and its subclasses are usually instantiated via
    `imitation.data.rollout.flatten_trajectories`.

    Indexing an instance `trans` of TransitionsMinimal with an integer `i`
    returns the `i`th `Dict[str, np.ndarray]` sample, whose keys are the field
    names of each dataclass field and whose values are the ith elements of each field
    value.

    Slicing returns a possibly empty instance of `TransitionsMinimal` where each
    field has been sliced.
    """

    obs: Observation
    """
    Previous observations. Shape: (batch_size, ) + observation_shape.

    The i'th observation `obs[i]` in this array is the observation seen
    by the agent when choosing action `acts[i]`. `obs[i]` is not required to
    be from the timestep preceding `obs[i+1]`.
    """

    acts: np.ndarray
    """Actions. Shape: (batch_size,) + action_shape."""

    infos: np.ndarray
    """Array of info dicts. Shape: (batch_size,)."""

    def __len__(self) -> int:
        """Returns number of transitions. Always positive."""
        return len(self.obs)

    def __post_init__(self):
        """Performs input validation: check shapes & dtypes match docstring.

        Also make array values read-only.

        Raises:
            ValueError: if batch size (array length) is inconsistent
                between `obs`, `acts` and `infos`.
        """
        for val in vars(self).values():
            if isinstance(val, np.ndarray):
                val.setflags(write=False)

        if len(self.obs) != len(self.acts):
            raise ValueError(
                "obs and acts must have same number of timesteps: "
                f"{len(self.obs)} != {len(self.acts)}",
            )

        if len(self.infos) != len(self.obs):
            raise ValueError(
                "obs and infos must have same number of timesteps: "
                f"{len(self.obs)} != {len(self.infos)}",
            )

    # TODO(adam): uncomment below once pytype bug fixed in
    # issue https://github.com/google/pytype/issues/1108
    # @overload
    # def __getitem__(self: T, key: slice) -> T:
    #     pass  # pragma: no cover
    #
    # @overload
    # def __getitem__(self, key: int) -> Mapping[str, np.ndarray]:
    #     pass  # pragma: no cover

    @overload
    def __getitem__(self, key: int) -> Mapping[str, np.ndarray]:
        pass

    @overload
    def __getitem__(self: TransitionsMinimalSelf, key: slice) -> TransitionsMinimalSelf:
        pass

    def __getitem__(self, key):
        """See TransitionsMinimal docstring for indexing and slicing semantics."""
        d = dataclass_quick_asdict(self)
        d_item = {k: v[key] for k, v in d.items()}

        if isinstance(key, slice):
            # Return type is the same as this dataclass. Replace field value with
            # slices.
            return dataclasses.replace(self, **d_item)
        else:
            assert isinstance(key, int)
            # Return type is a dictionary. Array values have no batch dimension.
            #
            # Dictionary of np.ndarray values is a convenient
            # torch.util.data.Dataset return type, as a torch.util.data.DataLoader
            # taking in this `Dataset` as its first argument knows how to
            # automatically concatenate several dictionaries together to make
            # a single dictionary batch with `torch.Tensor` values.
            return d_item


@dataclasses.dataclass(frozen=True)
class Transitions(TransitionsMinimal):
    """A batch of obs-act-obs-done transitions."""

    next_obs: Observation
    """New observation. Shape: (batch_size, ) + observation_shape.

    The i'th observation `next_obs[i]` in this array is the observation
    after the agent has taken action `acts[i]`.

    Invariants:
        * `next_obs.dtype == obs.dtype`
        * `len(next_obs) == len(obs)`
    """

    dones: np.ndarray
    """
    Boolean array indicating episode termination. Shape: (batch_size, ).

    `done[i]` is true iff `next_obs[i]` the last observation of an episode.
    """

    def __post_init__(self):
        """Performs input validation: check shapes & dtypes match docstring."""
        super().__post_init__()
        if self.obs.shape != self.next_obs.shape:
            raise ValueError(
                "obs and next_obs must have same shape: "
                f"{self.obs.shape} != {self.next_obs.shape}",
            )
        if self.obs.dtype != self.next_obs.dtype:
            raise ValueError(
                "obs and next_obs must have the same dtype: "
                f"{self.obs.dtype} != {self.next_obs.dtype}",
            )
        if self.dones.shape != (len(self.acts),):
            raise ValueError(
                "dones must be 1D array, one entry for each timestep: "
                f"{self.dones.shape} != ({len(self.acts)},)",
            )
        if self.dones.dtype != bool:
            raise ValueError(f"dones must be boolean, not {self.dones.dtype}")


@dataclasses.dataclass(frozen=True)
class TransitionsWithRew(Transitions):
    """A batch of obs-act-obs-rew-done transitions."""

    rews: np.ndarray
    """
    Reward. Shape: (batch_size, ). dtype float.

    The reward `rew[i]` at the i'th timestep is received after the
    agent has taken action `acts[i]`.
    """

    def __post_init__(self):
        """Performs input validation, including for rews."""
        super().__post_init__()
        _rews_validation(self.rews, self.acts)



"""DAgger (https://arxiv.org/pdf/1011.0686.pdf).

Interactively trains policy by collecting some demonstrations, doing BC, collecting more
demonstrations, doing BC again, etc. Initially the demonstrations just come from the
expert's policy; over time, they shift to be drawn more and more from the imitator's
policy.
"""

import abc
import logging
import os
import pathlib
import uuid
from typing import Any, Callable, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import torch as th
from stable_baselines3.common import policies, utils, vec_env
from stable_baselines3.common.vec_env.base_vec_env import VecEnvStepReturn
from torch.utils import data as th_data

from imitation.algorithms import base, bc
from imitation.data import rollout, serialize, types
from imitation.util import logger as imit_logger
from imitation.util import util


class BetaSchedule(abc.ABC):
    """Computes beta (% of time demonstration action used) from training round."""

    @abc.abstractmethod
    def __call__(self, round_num: int) -> float:
        """Computes the value of beta for the current round.

        Args:
            round_num: the current round number. Rounds are assumed to be sequentially
                numbered from 0.

        Returns:
            The fraction of the time to sample a demonstrator action. Robot
                actions will be sampled the remainder of the time.
        """  # noqa: DAR202


class LinearBetaSchedule(BetaSchedule):
    """Linearly-decreasing schedule for beta."""

    def __init__(self, rampdown_rounds: int) -> None:
        """Builds LinearBetaSchedule.

        Args:
            rampdown_rounds: number of rounds over which to anneal beta.
        """
        self.rampdown_rounds = rampdown_rounds

    def __call__(self, round_num: int) -> float:
        """Computes beta value.

        Args:
            round_num: the current round number.

        Returns:
            beta linearly decreasing from `1` to `0` between round `0` and
            `self.rampdown_rounds`. After that, it is 0.
        """
        assert round_num >= 0
        return min(1, max(0, (self.rampdown_rounds - round_num) / self.rampdown_rounds))


class ExponentialBetaSchedule(BetaSchedule):
    """Exponentially decaying schedule for beta."""

    def __init__(self, decay_probability: float):
        """Builds ExponentialBetaSchedule.

        Args:
            decay_probability: the decay factor for beta.

        Raises:
            ValueError: if `decay_probability` not within (0, 1].
        """
        if not (0 < decay_probability <= 1):
            raise ValueError("decay_probability lies outside the range (0, 1].")
        self.decay_probability = decay_probability

    def __call__(self, round_num: int) -> float:
        """Computes beta value.

        Args:
            round_num: the current round number.

        Returns:
            beta as `self.decay_probability ^ round_num`
        """
        assert round_num >= 0
        return self.decay_probability**round_num


def reconstruct_trainer(
    scratch_dir: types.AnyPath,
    venv: vec_env.VecEnv,
    custom_logger: Optional[imit_logger.HierarchicalLogger] = None,
    device: Union[th.device, str] = "auto",
) -> "DAggerTrainer":
    """Reconstruct trainer from the latest snapshot in some working directory.

    Requires vectorized environment and (optionally) a logger, as these objects
    cannot be serialized.

    Args:
        scratch_dir: path to the working directory created by a previous run of
            this algorithm. The directory should contain `checkpoint-latest.pt` and
            `policy-latest.pt` files.
        venv: Vectorized training environment.
        custom_logger: Where to log to; if None (default), creates a new logger.
        device: device on which to load the trainer.

    Returns:
        A deserialized `DAggerTrainer`.
    """
    custom_logger = custom_logger or imit_logger.configure()
    scratch_dir = util.parse_path(scratch_dir)
    checkpoint_path = scratch_dir / "checkpoint-latest.pt"
    trainer = th.load(checkpoint_path, map_location=utils.get_device(device))
    trainer.venv = venv
    trainer._logger = custom_logger
    return trainer


def _save_dagger_demo(
    trajectory: types.Trajectory,
    trajectory_index: int,
    save_dir: types.AnyPath,
    rng: np.random.Generator,
    prefix: str = "",
) -> None:
    save_dir = util.parse_path(save_dir)
    assert isinstance(trajectory, types.Trajectory)
    actual_prefix = f"{prefix}-" if prefix else ""
    randbits = int.from_bytes(rng.bytes(16), "big")
    random_uuid = uuid.UUID(int=randbits, version=4).hex
    filename = f"{actual_prefix}dagger-demo-{trajectory_index}-{random_uuid}.npz"
    npz_path = save_dir / filename
    assert (
        not npz_path.exists()
    ), "The following DAgger demonstration path already exists: {0}".format(npz_path)
    serialize.save(npz_path, [trajectory])
    logging.info(f"Saved demo at '{npz_path}'")


class InteractiveTrajectoryCollector(vec_env.VecEnvWrapper):
    """DAgger VecEnvWrapper for querying and saving expert actions.

    Every call to `.step(actions)` accepts and saves expert actions to `self.save_dir`,
    but only forwards expert actions to the wrapped VecEnv with probability
    `self.beta`. With probability `1 - self.beta`, a "robot" action (i.e
    an action from the imitation policy) is forwarded instead.

    Demonstrations are saved as `TrajectoryWithRew` to `self.save_dir` at the end
    of every episode.
    """

    traj_accum: Optional[rollout.TrajectoryAccumulator]
    _last_obs: Optional[np.ndarray]
    _last_user_actions: Optional[np.ndarray]

    def __init__(
        self,
        venv: vec_env.VecEnv,
        get_robot_acts: Callable[[np.ndarray], np.ndarray],
        beta: float,
        save_dir: types.AnyPath,
        rng: np.random.Generator,
    ) -> None:
        """Builds InteractiveTrajectoryCollector.

        Args:
            venv: vectorized environment to sample trajectories from.
            get_robot_acts: get robot actions that can be substituted for
                human actions. Takes a vector of observations as input & returns a
                vector of actions.
            beta: fraction of the time to use action given to .step() instead of
                robot action. The choice of robot or human action is independently
                randomized for each individual `Env` at every timestep.
            save_dir: directory to save collected trajectories in.
            rng: random state for random number generation.
        """
        super().__init__(venv)
        self.get_robot_acts = get_robot_acts
        assert 0 <= beta <= 1
        self.beta = beta
        self.traj_accum = None
        self.save_dir = save_dir
        self._last_obs = None
        self._done_before = True
        self._is_reset = False
        self._last_user_actions = None
        self.rng = rng

    def seed(self, seed: Optional[int] = None) -> List[Optional[int]]:
        """Set the seed for the DAgger random number generator and wrapped VecEnv.

        The DAgger RNG is used along with `self.beta` to determine whether the expert
        or robot action is forwarded to the wrapped VecEnv.

        Args:
            seed: The random seed. May be None for completely random seeding.

        Returns:
            A list containing the seeds for each individual env. Note that all list
            elements may be None, if the env does not return anything when seeded.
        """
        self.rng = np.random.default_rng(seed=seed)
        return list(self.venv.seed(seed))

    def reset(self) -> np.ndarray:
        """Resets the environment.

        Returns:
            obs: first observation of a new trajectory.
        """
        self.traj_accum = rollout.TrajectoryAccumulator()
        obs = self.venv.reset()
        assert isinstance(obs, np.ndarray)
        for i, ob in enumerate(obs):
            self.traj_accum.add_step({"obs": ob}, key=i)
        self._last_obs = obs
        self._is_reset = True
        self._last_user_actions = None
        return obs

    def step_async(self, actions: np.ndarray) -> None:
        """Steps with a `1 - beta` chance of using `self.get_robot_acts` instead.

        DAgger needs to be able to inject imitation policy actions randomly at some
        subset of time steps. This method has a `self.beta` chance of keeping the
        `actions` passed in as an argument, and a `1 - self.beta` chance of
        forwarding actions generated by `self.get_robot_acts` instead.
        "robot" (i.e. imitation policy) action if necessary.

        At the end of every episode, a `TrajectoryWithRew` is saved to `self.save_dir`,
        where every saved action is the expert action, regardless of whether the
        robot action was used during that timestep.

        Args:
            actions: the _intended_ demonstrator/expert actions for the current
                state. This will be executed with probability `self.beta`.
                Otherwise, a "robot" (typically a BC policy) action will be sampled
                and executed instead via `self.get_robot_act`.
        """
        assert self._is_reset, "call .reset() before .step()"
        assert self._last_obs is not None

        # Replace each given action with a robot action 100*(1-beta)% of the time.
        actual_acts = np.array(actions)

        mask = self.rng.uniform(0, 1, size=(self.num_envs,)) > self.beta
        if np.sum(mask) != 0:
            actual_acts[mask] = self.get_robot_acts(self._last_obs[mask])

        self._last_user_actions = actions
        self.venv.step_async(actual_acts)

    def step_wait(self) -> VecEnvStepReturn:
        """Returns observation, reward, etc after previous `step_async()` call.

        Stores the transition, and saves trajectory as demo once complete.

        Returns:
            Observation, reward, dones (is terminal?) and info dict.
        """
        next_obs, rews, dones, infos = self.venv.step_wait()
        assert isinstance(next_obs, np.ndarray)
        assert self.traj_accum is not None
        assert self._last_user_actions is not None
        self._last_obs = next_obs
        fresh_demos = self.traj_accum.add_steps_and_auto_finish(
            obs=next_obs,
            acts=self._last_user_actions,
            rews=rews,
            infos=infos,
            dones=dones,
        )
        for traj_index, traj in enumerate(fresh_demos):
            _save_dagger_demo(traj, traj_index, self.save_dir, self.rng)

        return next_obs, rews, dones, infos


class NeedsDemosException(Exception):
    """Signals demos need to be collected for current round before continuing."""


class DAggerTrainer(base.BaseImitationAlgorithm):
    """DAgger training class with low-level API suitable for interactive human feedback.

    In essence, this is just BC with some helpers for incrementally
    resuming training and interpolating between demonstrator/learnt policies.
    Interaction proceeds in "rounds" in which the demonstrator first provides a
    fresh set of demonstrations, and then an underlying `BC` is invoked to
    fine-tune the policy on the entire set of demonstrations collected in all
    rounds so far. Demonstrations and policy/trainer checkpoints are stored in a
    directory with the following structure::

       scratch-dir-name/
           checkpoint-001.pt
           checkpoint-002.pt
           …
           checkpoint-XYZ.pt
           checkpoint-latest.pt
           demos/
               round-000/
                   demos_round_000_000.npz
                   demos_round_000_001.npz
                   …
               round-001/
                   demos_round_001_000.npz
                   …
               …
               round-XYZ/
                   …
    """

    _all_demos: List[types.Trajectory]

    DEFAULT_N_EPOCHS: int = 4
    """The default number of BC training epochs in `extend_and_update`."""

    def __init__(
        self,
        *,
        venv: vec_env.VecEnv,
        scratch_dir: types.AnyPath,
        rng: np.random.Generator,
        beta_schedule: Optional[Callable[[int], float]] = None,
        bc_trainer: bc.BC,
        custom_logger: Optional[imit_logger.HierarchicalLogger] = None,
    ):
        """Builds DAggerTrainer.

        Args:
            venv: Vectorized training environment.
            scratch_dir: Directory to use to store intermediate training
                information (e.g. for resuming training).
            rng: random state for random number generation.
            beta_schedule: Provides a value of `beta` (the probability of taking
                expert action in any given state) at each round of training. If
                `None`, then `linear_beta_schedule` will be used instead.
            bc_trainer: A `BC` instance used to train the underlying policy.
            custom_logger: Where to log to; if None (default), creates a new logger.
        """
        super().__init__(custom_logger=custom_logger)

        if beta_schedule is None:
            beta_schedule = LinearBetaSchedule(15)
        self.beta_schedule = beta_schedule
        self.scratch_dir = util.parse_path(scratch_dir)
        self.venv = venv
        self.round_num = 0
        self._last_loaded_round = -1
        self._all_demos = []
        self.rng = rng

        utils.check_for_correct_spaces(
            self.venv,
            bc_trainer.observation_space,
            bc_trainer.action_space,
        )
        self.bc_trainer = bc_trainer
        self.bc_trainer.logger = self.logger

    def __getstate__(self):
        """Return state excluding non-pickleable objects."""
        d = dict(self.__dict__)
        del d["venv"]
        del d["_logger"]
        return d

    @property
    def logger(self) -> imit_logger.HierarchicalLogger:
        """Returns logger for this object."""
        return super().logger

    @logger.setter
    def logger(self, value: imit_logger.HierarchicalLogger) -> None:
        # DAgger and inner-BC logger should stay in sync
        self._logger = value
        self.bc_trainer.logger = value

    @property
    def policy(self) -> policies.BasePolicy:
        return self.bc_trainer.policy

    @property
    def batch_size(self) -> int:
        return self.bc_trainer.batch_size

    def _load_all_demos(self) -> Tuple[types.Transitions, List[int]]:
        num_demos_by_round = []
        for round_num in range(self._last_loaded_round + 1, self.round_num + 1):
            round_dir = self._demo_dir_path_for_round(round_num)
            demo_paths = self._get_demo_paths(round_dir)
            self._all_demos.extend(serialize.load(p)[0] for p in demo_paths)
            num_demos_by_round.append(len(demo_paths))
        logging.info(f"Loaded {len(self._all_demos)} total")
        demo_transitions = rollout.flatten_trajectories(self._all_demos)
        return demo_transitions, num_demos_by_round

    def _get_demo_paths(self, round_dir: pathlib.Path) -> List[pathlib.Path]:
        # listdir returns filenames in an arbitrary order that depends on the
        # file system implementation:
        # https://stackoverflow.com/questions/31534583/is-os-listdir-deterministic
        # To ensure the order is consistent across file systems,
        # we sort by the filename.
        filenames = sorted(os.listdir(round_dir))
        return [round_dir / f for f in filenames if f.endswith(".npz")]

    def _demo_dir_path_for_round(self, round_num: Optional[int] = None) -> pathlib.Path:
        if round_num is None:
            round_num = self.round_num
        return self.scratch_dir / "demos" / f"round-{round_num:03d}"

    def _try_load_demos(self) -> None:
        """Load the dataset for this round into self.bc_trainer as a DataLoader."""
        demo_dir = self._demo_dir_path_for_round()
        demo_paths = self._get_demo_paths(demo_dir) if demo_dir.is_dir() else []
        if len(demo_paths) == 0:
            raise NeedsDemosException(
                f"No demos found for round {self.round_num} in dir '{demo_dir}'. "
                f"Maybe you need to collect some demos? See "
                f".create_trajectory_collector()",
            )

        if self._last_loaded_round < self.round_num:
            transitions, num_demos = self._load_all_demos()
            logging.info(
                f"Loaded {sum(num_demos)} new demos from {len(num_demos)} rounds",
            )
            if len(transitions) < self.batch_size:
                raise ValueError(
                    "Not enough transitions to form a single batch: "
                    f"self.batch_size={self.batch_size} > "
                    f"len(transitions)={len(transitions)}",
                )
            data_loader = th_data.DataLoader(
                transitions,
                self.batch_size,
                drop_last=True,
                shuffle=True,
                collate_fn=types.transitions_collate_fn,
            )
            self.bc_trainer.set_demonstrations(data_loader)
            self._last_loaded_round = self.round_num

    def extend_and_update(
        self,
        bc_train_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> int:
        """Extend internal batch of data and train BC.

        Specifically, this method will load new transitions (if necessary), train
        the model for a while, and advance the round counter. If there are no fresh
        demonstrations in the demonstration directory for the current round, then
        this will raise a `NeedsDemosException` instead of training or advancing
        the round counter. In that case, the user should call
        `.create_trajectory_collector()` and use the returned
        `InteractiveTrajectoryCollector` to produce a new set of demonstrations for
        the current interaction round.

        Arguments:
            bc_train_kwargs: Keyword arguments for calling `BC.train()`. If
                the `log_rollouts_venv` key is not provided, then it is set to
                `self.venv` by default. If neither of the `n_epochs` and `n_batches`
                keys are provided, then `n_epochs` is set to `self.DEFAULT_N_EPOCHS`.

        Returns:
            New round number after advancing the round counter.
        """
        if bc_train_kwargs is None:
            bc_train_kwargs = {}
        else:
            bc_train_kwargs = dict(bc_train_kwargs)

        user_keys = bc_train_kwargs.keys()
        if "log_rollouts_venv" not in user_keys:
            bc_train_kwargs["log_rollouts_venv"] = self.venv

        if "n_epochs" not in user_keys and "n_batches" not in user_keys:
            bc_train_kwargs["n_epochs"] = self.DEFAULT_N_EPOCHS

        logging.info("Loading demonstrations")
        self._try_load_demos()
        logging.info(f"Training at round {self.round_num}")
        self.bc_trainer.train(**bc_train_kwargs)
        self.round_num += 1
        logging.info(f"New round number is {self.round_num}")
        return self.round_num

    def create_trajectory_collector(self) -> InteractiveTrajectoryCollector:
        """Create trajectory collector to extend current round's demonstration set.

        Returns:
            A collector configured with the appropriate beta, imitator policy, etc.
            for the current round. Refer to the documentation for
            `InteractiveTrajectoryCollector` to see how to use this.
        """
        save_dir = self._demo_dir_path_for_round()
        beta = self.beta_schedule(self.round_num)
        collector = InteractiveTrajectoryCollector(
            venv=self.venv,
            get_robot_acts=lambda acts: self.bc_trainer.policy.predict(acts)[0],
            beta=beta,
            save_dir=save_dir,
            rng=self.rng,
        )
        return collector

    def save_trainer(self) -> Tuple[pathlib.Path, pathlib.Path]:
        """Create a snapshot of trainer in the scratch/working directory.

        The created snapshot can be reloaded with `reconstruct_trainer()`.
        In addition to saving one copy of the policy in the trainer snapshot, this
        method saves a second copy of the policy in its own file. Having a second copy
        of the policy is convenient because it can be loaded on its own and passed to
        evaluation routines for other algorithms.

        Returns:
            checkpoint_path: a path to one of the created `DAggerTrainer` checkpoints.
            policy_path: a path to one of the created `DAggerTrainer` policies.
        """
        self.scratch_dir.mkdir(parents=True, exist_ok=True)

        # save full trainer checkpoints
        checkpoint_paths = [
            self.scratch_dir / f"checkpoint-{self.round_num:03d}.pt",
            self.scratch_dir / "checkpoint-latest.pt",
        ]
        for checkpoint_path in checkpoint_paths:
            th.save(self, checkpoint_path)

        # save policies separately for convenience
        policy_paths = [
            self.scratch_dir / f"policy-{self.round_num:03d}.pt",
            self.scratch_dir / "policy-latest.pt",
        ]
        for policy_path in policy_paths:
            util.save_policy(self.policy, policy_path)

        return checkpoint_paths[0], policy_paths[0]


class SimpleDAggerTrainer(DAggerTrainer):
    """Simpler subclass of DAggerTrainer for training with synthetic feedback."""

    def __init__(
        self,
        *,
        venv: vec_env.VecEnv,
        scratch_dir: types.AnyPath,
        expert_policy: policies.BasePolicy,
        rng: np.random.Generator,
        expert_trajs: Optional[Sequence[types.Trajectory]] = None,
        **dagger_trainer_kwargs,
    ):
        """Builds SimpleDAggerTrainer.

        Args:
            venv: Vectorized training environment. Note that when the robot
                action is randomly injected (in accordance with `beta_schedule`
                argument), every individual environment will get a robot action
                simultaneously for that timestep.
            scratch_dir: Directory to use to store intermediate training
                information (e.g. for resuming training).
            expert_policy: The expert policy used to generate synthetic demonstrations.
            rng: Random state to use for the random number generator.
            expert_trajs: Optional starting dataset that is inserted into the round 0
                dataset.
            dagger_trainer_kwargs: Other keyword arguments passed to the
                superclass initializer `DAggerTrainer.__init__`.

        Raises:
            ValueError: The observation or action space does not match between
                `venv` and `expert_policy`.
        """
        super().__init__(
            venv=venv,
            scratch_dir=scratch_dir,
            rng=rng,
            **dagger_trainer_kwargs,
        )
        self.expert_policy = expert_policy
        if expert_policy.observation_space != self.venv.observation_space:
            raise ValueError(
                "Mismatched observation space between expert_policy and venv",
            )
        if expert_policy.action_space != self.venv.action_space:
            raise ValueError("Mismatched action space between expert_policy and venv")

        # TODO(shwang):
        #   Might welcome Transitions and DataLoaders as sources of expert data
        #   in the future too, but this will require some refactoring, so for
        #   now we just have `expert_trajs`.
        if expert_trajs is not None:
            # Save each initial expert trajectory into the "round 0" demonstration
            # data directory.
            for traj_index, traj in enumerate(expert_trajs):
                _save_dagger_demo(
                    traj,
                    traj_index,
                    self._demo_dir_path_for_round(),
                    self.rng,
                    prefix="initial_data",
                )

    def train(
        self,
        total_timesteps: int,
        *,
        rollout_round_min_episodes: int = 3,
        rollout_round_min_timesteps: int = 500,
        bc_train_kwargs: Optional[dict] = None,
    ) -> None:
        """Train the DAgger agent.

        The agent is trained in "rounds" where each round consists of a dataset
        aggregation step followed by BC update step.

        During a dataset aggregation step, `self.expert_policy` is used to perform
        rollouts in the environment but there is a `1 - beta` chance (beta is
        determined from the round number and `self.beta_schedule`) that the DAgger
        agent's action is used instead. Regardless of whether the DAgger agent's action
        is used during the rollout, the expert action and corresponding observation are
        always appended to the dataset. The number of environment steps in the
        dataset aggregation stage is determined by the `rollout_round_min*` arguments.

        During a BC update step, `BC.train()` is called to update the DAgger agent on
        all data collected so far.

        Args:
            total_timesteps: The number of timesteps to train inside the environment.
                In practice this is a lower bound, because the number of timesteps is
                rounded up to finish the minimum number of episodes or timesteps in the
                last DAgger training round, and the environment timesteps are executed
                in multiples of `self.venv.num_envs`.
            rollout_round_min_episodes: The number of episodes the must be completed
                completed before a dataset aggregation step ends.
            rollout_round_min_timesteps: The number of environment timesteps that must
                be completed before a dataset aggregation step ends. Also, that any
                round will always train for at least `self.batch_size` timesteps,
                because otherwise BC could fail to receive any batches.
            bc_train_kwargs: Keyword arguments for calling `BC.train()`. If
                the `log_rollouts_venv` key is not provided, then it is set to
                `self.venv` by default. If neither of the `n_epochs` and `n_batches`
                keys are provided, then `n_epochs` is set to `self.DEFAULT_N_EPOCHS`.
        """
        total_timestep_count = 0
        round_num = 0

        while total_timestep_count < total_timesteps:
            collector = self.create_trajectory_collector()
            round_episode_count = 0
            round_timestep_count = 0

            sample_until = rollout.make_sample_until(
                min_timesteps=max(rollout_round_min_timesteps, self.batch_size),
                min_episodes=rollout_round_min_episodes,
            )

            trajectories = rollout.generate_trajectories(
                policy=self.expert_policy,
                venv=collector,
                sample_until=sample_until,
                deterministic_policy=True,
                rng=collector.rng,
            )

            for traj in trajectories:
                self._logger.record_mean(
                    "dagger/mean_episode_reward",
                    np.sum(traj.rews),
                )
                round_timestep_count += len(traj)
                total_timestep_count += len(traj)

            round_episode_count += len(trajectories)

            self._logger.record("dagger/total_timesteps", total_timestep_count)
            self._logger.record("dagger/round_num", round_num)
            self._logger.record("dagger/round_episode_count", round_episode_count)
            self._logger.record("dagger/round_timestep_count", round_timestep_count)

            # `logger.dump` is called inside BC.train within the following fn call:
            self.extend_and_update(bc_train_kwargs)
            round_num += 1



"""Config files for tuning experiments."""

import ray.tune as tune
import sacred
from torch import nn

from imitation.algorithms import dagger as dagger_alg
from imitation.scripts.parallel import parallel_ex

tuning_ex = sacred.Experiment("tuning", ingredients=[parallel_ex])


@tuning_ex.named_config
def rl():
    parallel_run_config = dict(
        sacred_ex_name="train_rl",
        run_name="rl_tuning",
        base_named_configs=["logging.wandb_logging"],
        base_config_updates={"environment": {"num_vec": 1}},
        search_space={
            "config_updates": {
                "rl": {
                    "batch_size": tune.choice([512, 1024, 2048, 4096, 8192]),
                    "rl_kwargs": {
                        "learning_rate": tune.loguniform(1e-5, 1e-2),
                        "batch_size": tune.choice([64, 128, 256, 512]),
                        "n_epochs": tune.choice([5, 10, 20]),
                    },
                },
            },
        },
        num_samples=100,
        repeat=1,
        resources_per_trial=dict(cpu=1),
    )
    num_eval_seeds = 5


@tuning_ex.named_config
def bc():
    parallel_run_config = dict(
        sacred_ex_name="train_imitation",
        run_name="bc_tuning",
        base_named_configs=["logging.wandb_logging"],
        base_config_updates={
            "environment": {"num_vec": 1},
            "demonstrations": {"source": "huggingface"},
        },
        search_space={
            "config_updates": {
                "bc": dict(
                    batch_size=tune.choice([8, 16, 32, 64]),
                    l2_weight=tune.loguniform(1e-6, 1e-2),  # L2 regularization weight
                    optimizer_kwargs=dict(
                        lr=tune.loguniform(1e-5, 1e-2),
                    ),
                    train_kwargs=dict(
                        n_epochs=tune.choice([1, 5, 10, 20]),
                    ),
                ),
            },
            "command_name": "bc",
        },
        num_samples=64,
        repeat=3,
        resources_per_trial=dict(cpu=1),
    )

    num_eval_seeds = 5
    eval_best_trial_resource_multiplier = 1


@tuning_ex.named_config
def dagger():
    parallel_run_config = dict(
        sacred_ex_name="train_imitation",
        run_name="dagger_tuning",
        base_named_configs=["logging.wandb_logging"],
        base_config_updates={
            "environment": {"num_vec": 1},
            "demonstrations": {"source": "huggingface"},
            "dagger": {"total_timesteps": 1e5},
            "bc": {
                "batch_size": 16,
                "l2_weight": 1e-4,
                "optimizer_kwargs": {"lr": 1e-3},
            },
        },
        search_space={
            "config_updates": {
                "bc": dict(
                    train_kwargs=dict(
                        n_epochs=tune.choice([1, 5, 10]),
                    ),
                ),
                "dagger": dict(
                    beta_schedule=tune.choice(
                        [dagger_alg.LinearBetaSchedule(i) for i in [1, 5, 15]]
                        + [
                            dagger_alg.ExponentialBetaSchedule(i)
                            for i in [0.3, 0.5, 0.7]
                        ],
                    ),
                    rollout_round_min_episodes=tune.choice([3, 5, 10]),
                ),
            },
            "command_name": "dagger",
        },
        num_samples=50,
        repeat=3,
        resources_per_trial=dict(cpu=1),
    )
    num_eval_seeds = 5


@tuning_ex.named_config
def gail():
    parallel_run_config = dict(
        sacred_ex_name="train_adversarial",
        run_name="gail_tuning",
        base_named_configs=["logging.wandb_logging"],
        base_config_updates={
            "environment": {"num_vec": 1},
            "demonstrations": {"source": "huggingface"},
            "total_timesteps": 1e7,
        },
        search_space={
            "config_updates": {
                "algorithm_kwargs": dict(
                    demo_batch_size=tune.choice([32, 128, 512, 2048, 8192]),
                    n_disc_updates_per_round=tune.choice([8, 16]),
                ),
                "rl": {
                    "batch_size": tune.choice([4096, 8192, 16384]),
                    "rl_kwargs": {
                        "ent_coef": tune.loguniform(1e-7, 1e-3),
                        "learning_rate": tune.loguniform(1e-5, 1e-2),
                    },
                },
                "algorithm_specific": {},
            },
            "command_name": "gail",
        },
        num_samples=100,
        repeat=3,
        resources_per_trial=dict(cpu=1),
    )
    num_eval_seeds = 5


@tuning_ex.named_config
def airl():
    parallel_run_config = dict(
        sacred_ex_name="train_adversarial",
        run_name="airl_tuning",
        base_named_configs=["logging.wandb_logging"],
        base_config_updates={
            "environment": {"num_vec": 1},
            "demonstrations": {"source": "huggingface"},
            "total_timesteps": 1e7,
        },
        search_space={
            "config_updates": {
                "algorithm_kwargs": dict(
                    demo_batch_size=tune.choice([32, 128, 512, 2048, 8192]),
                    n_disc_updates_per_round=tune.choice([8, 16]),
                ),
                "rl": {
                    "batch_size": tune.choice([4096, 8192, 16384]),
                    "rl_kwargs": {
                        "ent_coef": tune.loguniform(1e-7, 1e-3),
                        "learning_rate": tune.loguniform(1e-5, 1e-2),
                    },
                },
                "algorithm_specific": {},
            },
            "command_name": "airl",
        },
        num_samples=100,
        repeat=3,
        resources_per_trial=dict(cpu=1),
    )
    num_eval_seeds = 5


@tuning_ex.named_config
def pc():
    parallel_run_config = dict(
        sacred_ex_name="train_preference_comparisons",
        run_name="pc_tuning",
        base_named_configs=["logging.wandb_logging"],
        base_config_updates={
            "environment": {"num_vec": 1},
            "demonstrations": {"source": "huggingface"},
            "total_timesteps": 2e7,
            "total_comparisons": 5000,
            "query_schedule": "hyperbolic",
            "gatherer_kwargs": {"sample": True},
        },
        search_space={
            "named_configs": [
                ["reward.normalize_output_disable"],
            ],
            "config_updates": {
                "train": {
                    "policy_kwargs": {
                        "activation_fn": tune.choice(
                            [
                                nn.ReLU,
                            ],
                        ),
                    },
                },
                "num_iterations": tune.choice([25, 50]),
                "initial_comparison_frac": tune.choice([0.1, 0.25]),
                "reward_trainer_kwargs": {
                    "epochs": tune.choice([1, 3, 6]),
                },
                "rl": {
                    "batch_size": tune.choice([512, 2048, 8192]),
                    "rl_kwargs": {
                        "learning_rate": tune.loguniform(1e-5, 1e-2),
                        "ent_coef": tune.loguniform(1e-7, 1e-3),
                    },
                },
            },
        },
        num_samples=100,
        repeat=3,
        resources_per_trial=dict(cpu=1),
    )
    num_eval_seeds = 5



"""Wrapper to record rendered video frames from an environment."""

import pathlib
from typing import Any, Dict, Optional, SupportsFloat, Tuple

import gymnasium as gym
from gymnasium.core import WrapperActType, WrapperObsType
from gymnasium.wrappers.monitoring import video_recorder


class VideoWrapper(gym.Wrapper):
    """Creates videos from wrapped environment by calling render after each timestep."""

    episode_id: int
    video_recorder: Optional[video_recorder.VideoRecorder]
    single_video: bool
    directory: pathlib.Path

    def __init__(
        self,
        env: gym.Env,
        directory: pathlib.Path,
        single_video: bool = True,
    ):
        """Builds a VideoWrapper.

        Args:
            env: the wrapped environment.
            directory: the output directory.
            single_video: if True, generates a single video file, with episodes
                concatenated. If False, a new video file is created for each episode.
                Usually a single video file is what is desired. However, if one is
                searching for an interesting episode (perhaps by looking at the
                metadata), then saving to different files can be useful.
        """
        super().__init__(env)
        self.episode_id = 0
        self.video_recorder = None
        self.single_video = single_video

        self.directory = directory
        self.directory.mkdir(parents=True, exist_ok=True)

    def _reset_video_recorder(self) -> None:
        """Creates a video recorder if one does not already exist.

        Called at the start of each episode (by `reset`). When a video recorder is
        already present, it will only create a new one if `self.single_video == False`.
        """
        if self.video_recorder is not None:
            # Video recorder already started.
            if not self.single_video:
                # We want a new video for each episode, so destroy current recorder.
                self.video_recorder.close()
                self.video_recorder = None

        if self.video_recorder is None:
            # No video recorder -- start a new one.
            self.video_recorder = video_recorder.VideoRecorder(
                env=self.env,
                base_path=str(self.directory / f"video.{self.episode_id:06}"),
                metadata={"episode_id": self.episode_id},
            )

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> Tuple[WrapperObsType, Dict[str, Any]]:
        self._reset_video_recorder()
        self.episode_id += 1
        return super().reset(seed=seed, options=options)

    def step(
        self,
        action: WrapperActType,
    ) -> Tuple[WrapperObsType, SupportsFloat, bool, bool, Dict[str, Any]]:
        res = super().step(action)
        assert self.video_recorder is not None
        self.video_recorder.capture_frame()
        return res

    def close(self) -> None:
        if self.video_recorder is not None:
            self.video_recorder.close()
            self.video_recorder = None
        super().close()



"""Finite-horizon tabular Maximum Causal Entropy IRL.

Follows the description in chapters 9 and 10 of Brian Ziebart's `PhD thesis`_.

.. _PhD thesis:
    http://www.cs.cmu.edu/~bziebart/publications/thesis-bziebart.pdf
"""
import collections
import warnings
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    Tuple,
    Type,
    Union,
)

import gymnasium as gym
import numpy as np
import scipy.special
import torch as th
from seals import base_envs
from stable_baselines3.common import policies
from stable_baselines3.common import type_aliases as sb3_types

from imitation.algorithms import base
from imitation.data import rollout, types
from imitation.rewards import reward_nets
from imitation.util import logger as imit_logger
from imitation.util import networks, util


def mce_partition_fh(
    env: base_envs.TabularModelPOMDP,
    *,
    reward: Optional[np.ndarray] = None,
    discount: float = 1.0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""Performs the soft Bellman backup for a finite-horizon MDP.

    Calculates V^{soft}, Q^{soft}, and pi using recurrences (9.1), (9.2), and
    (9.3) from Ziebart (2010).

    Args:
        env: a tabular, known-dynamics MDP.
        reward: a reward matrix. Defaults to env.reward_matrix.
        discount: discount rate.

    Returns:
        (V, Q, \pi) corresponding to the soft values, Q-values and MCE policy.
        V is a 2d array, indexed V[t,s]. Q is a 3d array, indexed Q[t,s,a].
        \pi is a 3d array, indexed \pi[t,s,a].

    Raises:
        ValueError: if ``env.horizon`` is None (infinite horizon).
    """
    # shorthand
    horizon = env.horizon
    if horizon is None:
        raise ValueError("Only finite-horizon environments are supported.")
    n_states = env.state_dim
    n_actions = env.action_dim
    T = env.transition_matrix
    if reward is None:
        reward = env.reward_matrix

    # Initialization
    # indexed as V[t,s]
    V = np.full((horizon, n_states), -np.inf)
    # indexed as Q[t,s,a]
    Q = np.zeros((horizon, n_states, n_actions))
    broad_R = reward[:, None]

    # Base case: final timestep
    # final Q(s,a) is just reward
    Q[horizon - 1, :, :] = broad_R
    # V(s) is always normalising constant
    V[horizon - 1, :] = scipy.special.logsumexp(Q[horizon - 1, :, :], axis=1)

    # Recursive case
    for t in reversed(range(horizon - 1)):
        next_values_s_a = T @ V[t + 1, :]
        Q[t, :, :] = broad_R + discount * next_values_s_a
        V[t, :] = scipy.special.logsumexp(Q[t, :, :], axis=1)

    pi = np.exp(Q - V[:, :, None])

    return V, Q, pi


def mce_occupancy_measures(
    env: base_envs.TabularModelPOMDP,
    *,
    reward: Optional[np.ndarray] = None,
    pi: Optional[np.ndarray] = None,
    discount: float = 1.0,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate state visitation frequency Ds for each state s under a given policy pi.

    You can get pi from `mce_partition_fh`.

    Args:
        env: a tabular MDP.
        reward: reward matrix. Defaults is env.reward_matrix.
        pi: policy to simulate. Defaults to soft-optimal policy w.r.t reward
            matrix.
        discount: rate to discount the cumulative occupancy measure D.

    Returns:
        Tuple of ``D`` (ndarray) and ``Dcum`` (ndarray). ``D`` is of shape
        ``(env.horizon, env.n_states)`` and records the probability of being in a
        given state at a given timestep. ``Dcum`` is of shape ``(env.n_states,)``
        and records the expected discounted number of times each state is visited.

    Raises:
        ValueError: if ``env.horizon`` is None (infinite horizon).
    """
    # shorthand
    horizon = env.horizon
    if horizon is None:
        raise ValueError("Only finite-horizon environments are supported.")
    n_states = env.state_dim
    n_actions = env.action_dim
    T = env.transition_matrix
    if reward is None:
        reward = env.reward_matrix
    if pi is None:
        _, _, pi = mce_partition_fh(env, reward=reward)

    D = np.zeros((horizon + 1, n_states))
    D[0, :] = env.initial_state_dist
    for t in range(horizon):
        for a in range(n_actions):
            E = D[t] * pi[t, :, a]
            D[t + 1, :] += E @ T[:, a, :]

    Dcum = rollout.discounted_sum(D, discount)
    assert isinstance(Dcum, np.ndarray)
    return D, Dcum


def squeeze_r(r_output: th.Tensor) -> th.Tensor:
    """Squeeze a reward output tensor down to one dimension, if necessary.

    Args:
         r_output (th.Tensor): output of reward model. Can be either 1D
            ([n_states]) or 2D ([n_states, 1]).

    Returns:
         squeezed reward of shape [n_states].
    """
    if r_output.ndim == 2:
        return th.squeeze(r_output, 1)
    assert r_output.ndim == 1
    return r_output


class TabularPolicy(policies.BasePolicy):
    """A tabular policy. Cannot be trained -- prediction only."""

    pi: np.ndarray
    rng: np.random.Generator

    def __init__(
        self,
        state_space: gym.Space,
        action_space: gym.Space,
        pi: np.ndarray,
        rng: np.random.Generator,
    ) -> None:
        """Builds TabularPolicy.

        Args:
            state_space: The state space of the environment.
            action_space: The action space of the environment.
            pi: A tabular policy. Three-dimensional array, where pi[t,s,a]
                is the probability of taking action a at state s at timestep t.
            rng: Random state, used for sampling when `predict` is called with
                `deterministic=False`.
        """
        assert isinstance(state_space, gym.spaces.Discrete), "state not tabular"
        assert isinstance(action_space, gym.spaces.Discrete), "action not tabular"
        # What we call state space here is observation space in SB3 nomenclature.
        super().__init__(observation_space=state_space, action_space=action_space)
        self.rng = rng
        self.set_pi(pi)

    def set_pi(self, pi: np.ndarray) -> None:
        """Sets tabular policy to `pi`."""
        assert pi.ndim == 3, "expected three-dimensional policy"
        assert np.allclose(pi.sum(axis=2), 1), "policy not normalized"
        assert np.all(pi >= 0), "policy has negative probabilities"
        self.pi = pi

    def _predict(self, observation: sb3_types.PyTorchObs, deterministic: bool = False):
        raise NotImplementedError("Should never be called as predict overridden.")

    def forward(
        self,
        observation: th.Tensor,
        deterministic: bool = False,
    ) -> NoReturn:
        raise NotImplementedError("Should never be called.")  # pragma: no cover

    def predict(
        self,
        observation: Union[np.ndarray, Mapping[str, np.ndarray]],
        state: Optional[Tuple[np.ndarray, ...]] = None,
        episode_start: Optional[np.ndarray] = None,
        deterministic: bool = False,
    ) -> Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]]:
        """Predict action to take in given state.

        Arguments follow SB3 naming convention as this is an SB3 policy.
        In this convention, observations are returned by the environment,
        and state is a hidden state used by the policy (used by us to
        keep track of timesteps).

        What is `observation` here is a state in the underlying MDP,
        and would be called `state` elsewhere in this file.

        Args:
            observation: States in the underlying MDP.
            state: Hidden states of the policy -- used to represent timesteps by us.
            episode_start: Has episode completed?
            deterministic: If true, pick action with highest probability; otherwise,
                sample.

        Returns:
            Tuple of the actions and new hidden states.
        """
        if state is None:
            timesteps = np.zeros(len(observation), dtype=int)
        else:
            assert len(state) == 1
            timesteps = state[0]
        assert len(timesteps) == len(observation), "timestep and obs batch size differ"

        if episode_start is not None:
            timesteps[episode_start] = 0

        actions: List[int] = []
        for obs, t in zip(observation, timesteps):
            assert self.observation_space.contains(obs), "illegal state"
            dist = self.pi[t, obs, :]
            if deterministic:
                actions.append(int(dist.argmax()))
            else:
                actions.append(self.rng.choice(len(dist), p=dist))

        timesteps += 1  # increment timestep
        state = (timesteps,)
        return np.array(actions), state


MCEDemonstrations = Union[np.ndarray, base.AnyTransitions]


class MCEIRL(base.DemonstrationAlgorithm[types.TransitionsMinimal]):
    """Tabular MCE IRL.

    Reward is a function of observations, but policy is a function of states.

    The "observations" effectively exist just to let MCE IRL learn a reward
    in a reasonable feature space, giving a helpful inductive bias, e.g. that
    similar states have similar reward.

    Since we are performing planning to compute the policy, there is no need
    for function approximation in the policy.
    """

    demo_state_om: Optional[np.ndarray]

    def __init__(
        self,
        demonstrations: Optional[MCEDemonstrations],
        env: base_envs.TabularModelPOMDP,
        reward_net: reward_nets.RewardNet,
        rng: np.random.Generator,
        optimizer_cls: Type[th.optim.Optimizer] = th.optim.Adam,
        optimizer_kwargs: Optional[Mapping[str, Any]] = None,
        discount: float = 1.0,
        linf_eps: float = 1e-3,
        grad_l2_eps: float = 1e-4,
        # TODO(adam): do we need log_interval or can just use record_mean...?
        log_interval: Optional[int] = 100,
        *,
        custom_logger: Optional[imit_logger.HierarchicalLogger] = None,
    ) -> None:
        r"""Creates MCE IRL.

        Args:
            demonstrations: Demonstrations from an expert (optional). Can be a sequence
                of trajectories, or transitions, an iterable over mappings that
                represent a batch of transitions, or a state occupancy measure.
                The demonstrations must have observations one-hot coded unless
                demonstrations is a state-occupancy measure.
            env: a tabular MDP.
            rng: random state used for sampling from policy.
            reward_net: a neural network that computes rewards for the supplied
                observations.
            optimizer_cls: optimizer to use for supervised training.
            optimizer_kwargs: keyword arguments for optimizer construction.
            discount: the discount factor to use when computing occupancy measure.
                If not 1.0 (undiscounted), then `demonstrations` must either be
                a (discounted) state-occupancy measure, or trajectories. Transitions
                are *not allowed* as we cannot discount them appropriately without
                knowing the timestep they were drawn from.
            linf_eps: optimisation terminates if the $l_{\infty}$ distance between
                the demonstrator's state occupancy measure and the state occupancy
                measure for the current reward falls below this value.
            grad_l2_eps: optimisation also terminates if the $\ell_2$ norm of the
                MCE IRL gradient falls below this value.
            log_interval: how often to log current loss stats (using `logging`).
                None to disable.
            custom_logger: Where to log to; if None (default), creates a new logger.

        Raises:
            ValueError: if the env horizon is not finite (or an integer).
        """
        self.discount = discount
        self.env = env
        self.demo_state_om = None
        super().__init__(
            demonstrations=demonstrations,
            custom_logger=custom_logger,
        )

        self.reward_net = reward_net
        optimizer_kwargs = optimizer_kwargs or {"lr": 1e-2}
        self.optimizer = optimizer_cls(reward_net.parameters(), **optimizer_kwargs)

        self.linf_eps = linf_eps
        self.grad_l2_eps = grad_l2_eps
        self.log_interval = log_interval
        self.rng = rng

        # Initialize policy to be uniform random. We don't use this for MCE IRL
        # training, but it gives us something to return at all times with `policy`
        # property, similar to other algorithms.
        if self.env.horizon is None:
            raise ValueError("Only finite-horizon environments are supported.")
        ones = np.ones((self.env.horizon, self.env.state_dim, self.env.action_dim))
        uniform_pi = ones / self.env.action_dim
        self._policy = TabularPolicy(
            state_space=self.env.state_space,
            action_space=self.env.action_space,
            pi=uniform_pi,
            rng=self.rng,
        )

    def _set_demo_from_trajectories(self, trajs: Iterable[types.Trajectory]) -> None:
        self.demo_state_om = np.zeros((self.env.state_dim,))
        num_demos = 0
        for traj in trajs:
            cum_discount = 1.0
            for obs in types.assert_not_dictobs(traj.obs):
                self.demo_state_om[obs] += cum_discount
                cum_discount *= self.discount
            num_demos += 1
        self.demo_state_om /= num_demos

    def _set_demo_from_obs(
        self,
        obses: np.ndarray,
        dones: Optional[np.ndarray],
        next_obses: Optional[np.ndarray],
    ) -> None:
        self.demo_state_om = np.zeros((self.env.state_dim,))

        for obs in obses:
            if isinstance(obs, th.Tensor):
                obs = obs.item()  # must be scalar
            self.demo_state_om[obs] += 1.0

        # We assume the transitions were flattened from some trajectories,
        # then possibly shuffled. So add next observations for terminal states,
        # as they will not appear anywhere else; but ignore next observations
        # for all other states as they occur elsewhere in dataset.
        if dones is not None and next_obses is not None:
            for done, obs in zip(dones, next_obses):
                if isinstance(done, th.Tensor):
                    done = done.item()  # must be scalar
                    obs = obs.item()  # must be scalar
                if done:
                    self.demo_state_om[obs] += 1.0
        else:
            warnings.warn(
                "Training MCEIRL with transitions that lack next observation."
                "This will result in systematically wrong occupancy measure estimates.",
            )

        # Normalize occupancy measure estimates
        assert self.env.horizon is not None
        self.demo_state_om *= (self.env.horizon + 1) / self.demo_state_om.sum()

    def set_demonstrations(self, demonstrations: MCEDemonstrations) -> None:
        if isinstance(demonstrations, np.ndarray):
            # Demonstrations are an occupancy measure
            assert demonstrations.ndim == 1
            self.demo_state_om = demonstrations
            return

        # Demonstrations are either trajectories or transitions;
        # we must compute occupancy measure from this.
        if isinstance(demonstrations, Iterable):
            first_item, demonstrations_it = util.get_first_iter_element(demonstrations)
            if isinstance(first_item, types.Trajectory):
                self._set_demo_from_trajectories(demonstrations_it)
                return

        # Demonstrations are from some kind of transitions-like object. This does
        # not contain timesteps, so can only compute OM when undiscounted.
        if self.discount != 1.0:
            raise ValueError(
                "Cannot compute discounted OM from timeless Transitions.",
            )

        if isinstance(demonstrations, types.Transitions):
            self._set_demo_from_obs(
                types.assert_not_dictobs(demonstrations.obs),
                demonstrations.dones,
                types.assert_not_dictobs(demonstrations.next_obs),
            )
        elif isinstance(demonstrations, types.TransitionsMinimal):
            self._set_demo_from_obs(
                types.assert_not_dictobs(demonstrations.obs),
                None,
                None,
            )
        elif isinstance(demonstrations, Iterable):
            # Demonstrations are a Torch DataLoader or other Mapping iterable
            # Collect them together into one big NumPy array. This is inefficient,
            # we could compute the running statistics instead, but in practice do
            # not expect large dataset sizes together with MCE IRL.
            collated_list: Dict[
                str,
                List[types.AnyTensor],
            ] = collections.defaultdict(list)
            for batch in demonstrations:
                assert isinstance(batch, Mapping)
                for k in ("obs", "dones", "next_obs"):
                    x = batch.get(k)
                    if x is not None:
                        assert isinstance(x, (np.ndarray, th.Tensor))
                        collated_list[k].append(x)
            collated = {k: np.concatenate(v) for k, v in collated_list.items()}

            assert "obs" in collated
            for k, v in collated.items():
                assert len(v) == len(collated["obs"]), k
            self._set_demo_from_obs(
                collated["obs"],
                collated.get("dones"),
                collated.get("next_obs"),
            )
        else:
            raise TypeError(
                f"Unsupported demonstration type {type(demonstrations)}",
            )

    def _train_step(self, obs_mat: th.Tensor) -> Tuple[np.ndarray, np.ndarray]:
        self.optimizer.zero_grad()

        # get reward predicted for each state by current model, & compute
        # expected # of times each state is visited by soft-optimal policy
        # w.r.t that reward function
        # TODO(adam): support not just state-only reward?
        predicted_r = squeeze_r(self.reward_net(obs_mat, None, None, None))
        assert predicted_r.shape == (obs_mat.shape[0],)
        predicted_r_np = predicted_r.detach().cpu().numpy()
        _, visitations = mce_occupancy_measures(
            self.env,
            reward=predicted_r_np,
            discount=self.discount,
        )

        # Forward/back/step (grads are zeroed at the top).
        # weights_th(s) = \pi(s) - D(s)
        weights_th = th.as_tensor(
            visitations - self.demo_state_om,
            dtype=self.reward_net.dtype,
            device=self.reward_net.device,
        )
        # The "loss" is then:
        #   E_\pi[r_\theta(S)] - E_D[r_\theta(S)]
        loss = th.dot(weights_th, predicted_r)
        # This gives the required gradient:
        #   E_\pi[\nabla r_\theta(S)] - E_D[\nabla r_\theta(S)].
        loss.backward()
        self.optimizer.step()

        return predicted_r_np, visitations

    def train(self, max_iter: int = 1000) -> np.ndarray:
        """Runs MCE IRL.

        Args:
            max_iter: The maximum number of iterations to train for. May terminate
                earlier if `self.linf_eps` or `self.grad_l2_eps` thresholds are reached.

        Returns:
            State occupancy measure for the final reward function. `self.reward_net`
            and `self.optimizer` will be updated in-place during optimisation.
        """
        # use the same device and dtype as the rmodel parameters
        obs_mat = self.env.observation_matrix
        torch_obs_mat = th.as_tensor(
            obs_mat,
            dtype=self.reward_net.dtype,
            device=self.reward_net.device,
        )
        assert self.demo_state_om is not None
        assert self.demo_state_om.shape == (len(obs_mat),)

        with networks.training(self.reward_net):
            # switch to training mode (affects dropout, normalization)
            for t in range(max_iter):
                predicted_r_np, visitations = self._train_step(torch_obs_mat)

                # these are just for termination conditions & debug logging
                grads = []
                for p in self.reward_net.parameters():
                    assert p.grad is not None  # for type checker
                    grads.append(p.grad)
                grad_norm = util.tensor_iter_norm(grads).item()
                linf_delta = np.max(np.abs(self.demo_state_om - visitations))

                if self.log_interval is not None and 0 == (t % self.log_interval):
                    params = self.reward_net.parameters()
                    weight_norm = util.tensor_iter_norm(params).item()
                    self.logger.record("iteration", t)
                    self.logger.record("linf_delta", linf_delta)
                    self.logger.record("weight_norm", weight_norm)
                    self.logger.record("grad_norm", grad_norm)
                    self.logger.dump(t)

                if linf_delta <= self.linf_eps or grad_norm <= self.grad_l2_eps:
                    break

        _, _, pi = mce_partition_fh(
            self.env,
            reward=predicted_r_np,
            discount=self.discount,
        )
        # TODO(adam): this policy works on states, not observations, so can't
        # actually compute rollouts from it in the usual way. Fix this by making
        # observations part of MCE IRL and turn environment from POMDP->MDP?
        self._policy.set_pi(pi)

        return visitations

    @property
    def policy(self) -> policies.BasePolicy:
        return self._policy



"""Train a reward model using preference comparisons.

Can be used as a CLI script, or the `train_preference_comparisons` function
can be called directly.
"""

import functools
import pathlib
from typing import Any, Mapping, Optional, Type, Union

import numpy as np
import torch as th
from sacred.observers import FileStorageObserver
from stable_baselines3.common import type_aliases

import imitation.data.serialize as data_serialize
import imitation.policies.serialize as policies_serialize
from imitation.algorithms import preference_comparisons
from imitation.scripts.config.train_preference_comparisons import (
    train_preference_comparisons_ex,
)
from imitation.scripts.ingredients import environment
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients import policy_evaluation, reward
from imitation.scripts.ingredients import rl as rl_common


def save_model(
    agent_trainer: preference_comparisons.AgentTrainer,
    save_path: pathlib.Path,
):
    """Save the model as `model.zip`."""
    policies_serialize.save_stable_model(
        output_dir=save_path / "policy",
        model=agent_trainer.algorithm,
    )


def save_checkpoint(
    trainer: preference_comparisons.PreferenceComparisons,
    save_path: pathlib.Path,
    allow_save_policy: Optional[bool],
):
    """Save reward model and optionally policy."""
    save_path.mkdir(parents=True, exist_ok=True)
    th.save(trainer.model, save_path / "reward_net.pt")
    if allow_save_policy:
        # Note: We should only save the model as model.zip if `trajectory_generator`
        # contains one. Currently we are slightly over-conservative, by requiring
        # that an AgentTrainer be used if we're saving the policy.
        assert isinstance(
            trainer.trajectory_generator,
            preference_comparisons.AgentTrainer,
        )
        save_model(trainer.trajectory_generator, save_path)
    else:
        trainer.logger.warn(
            "trainer.trajectory_generator doesn't contain a policy to save.",
        )


@train_preference_comparisons_ex.main
def train_preference_comparisons(
    total_timesteps: int,
    total_comparisons: int,
    num_iterations: int,
    comparison_queue_size: Optional[int],
    fragment_length: int,
    transition_oversampling: float,
    initial_comparison_frac: float,
    exploration_frac: float,
    trajectory_path: Optional[str],
    trajectory_generator_kwargs: Mapping[str, Any],
    save_preferences: bool,
    agent_path: Optional[str],
    preference_model_kwargs: Mapping[str, Any],
    reward_trainer_kwargs: Mapping[str, Any],
    gatherer_cls: Type[preference_comparisons.PreferenceGatherer],
    gatherer_kwargs: Mapping[str, Any],
    active_selection: bool,
    active_selection_oversampling: int,
    uncertainty_on: str,
    fragmenter_kwargs: Mapping[str, Any],
    allow_variable_horizon: bool,
    checkpoint_interval: int,
    query_schedule: Union[str, type_aliases.Schedule],
    _rnd: np.random.Generator,
) -> Mapping[str, Any]:
    """Train a reward model using preference comparisons.

    Args:
        total_timesteps: number of environment interaction steps
        total_comparisons: number of preferences to gather in total
        num_iterations: number of times to train the agent against the reward model
            and then train the reward model against newly gathered preferences.
        comparison_queue_size: the maximum number of comparisons to keep in the
            queue for training the reward model. If None, the queue will grow
            without bound as new comparisons are added.
        fragment_length: number of timesteps per fragment that is used to elicit
            preferences
        transition_oversampling: factor by which to oversample transitions before
            creating fragments. Since fragments are sampled with replacement,
            this is usually chosen > 1 to avoid having the same transition
            in too many fragments.
        initial_comparison_frac: fraction of total_comparisons that will be
            sampled before the rest of training begins (using the randomly initialized
            agent). This can be used to pretrain the reward model before the agent
            is trained on the learned reward.
        exploration_frac: fraction of trajectory samples that will be created using
            partially random actions, rather than the current policy. Might be helpful
            if the learned policy explores too little and gets stuck with a wrong
            reward.
        trajectory_path: either None, in which case an agent will be trained
            and used to sample trajectories on the fly, or a path to a pickled
            sequence of TrajectoryWithRew to be trained on.
        trajectory_generator_kwargs: kwargs to pass to the trajectory generator.
        save_preferences: if True, store the final dataset of preferences to disk.
        agent_path: if given, initialize the agent using this stored policy
            rather than randomly.
        preference_model_kwargs: passed to PreferenceModel
        reward_trainer_kwargs: passed to BasicRewardTrainer or EnsembleRewardTrainer
        gatherer_cls: type of PreferenceGatherer to use (defaults to SyntheticGatherer)
        gatherer_kwargs: passed to the PreferenceGatherer specified by gatherer_cls
        active_selection: use active selection fragmenter instead of random fragmenter
        active_selection_oversampling: factor by which to oversample random fragments
            from the base fragmenter of active selection.
            this is usually chosen > 1 to allow the active selection algorithm to pick
            fragment pairs with highest uncertainty. = 1 implies no active selection.
        uncertainty_on: passed to ActiveSelectionFragmenter
        fragmenter_kwargs: passed to RandomFragmenter
        allow_variable_horizon: If False (default), algorithm will raise an
            exception if it detects trajectories of different length during
            training. If True, overrides this safety check. WARNING: variable
            horizon episodes leak information about the reward via termination
            condition, and can seriously confound evaluation. Read
            https://imitation.readthedocs.io/en/latest/guide/variable_horizon.html
            before overriding this.
        checkpoint_interval: Save the reward model and policy models (if
            trajectory_generator contains a policy) every `checkpoint_interval`
            iterations and after training is complete. If 0, then only save weights
            after training is complete. If <0, then don't save weights at all.
        query_schedule: one of ("constant", "hyperbolic", "inverse_quadratic").
            A function indicating how the total number of preference queries should
            be allocated to each iteration. "hyperbolic" and "inverse_quadratic"
            apportion fewer queries to later iterations when the policy is assumed
            to be better and more stable.
        _rnd: Random number generator provided by Sacred.

    Returns:
        Rollout statistics from trained policy.

    Raises:
        ValueError: Inconsistency between config and deserialized policy normalization.
    """
    # This allows to specify total_timesteps, total_comparisons etc. in scientific
    # notation, which is interpreted as a float by python.
    total_timesteps = int(total_timesteps)
    total_comparisons = int(total_comparisons)
    num_iterations = int(num_iterations)
    comparison_queue_size = (
        int(comparison_queue_size) if comparison_queue_size is not None else None
    )
    fragment_length = int(fragment_length)
    active_selection_oversampling = int(active_selection_oversampling)
    checkpoint_interval = int(checkpoint_interval)

    custom_logger, log_dir = logging_ingredient.setup_logging()

    with environment.make_venv() as venv:  # type: ignore[wrong-arg-count]
        reward_net = reward.make_reward_net(venv)
        relabel_reward_fn = functools.partial(
            reward_net.predict_processed,
            update_stats=False,
        )
        if agent_path is None:
            agent = rl_common.make_rl_algo(venv, relabel_reward_fn=relabel_reward_fn)
        else:
            agent = rl_common.load_rl_algo_from_path(
                agent_path=agent_path,
                venv=venv,
                relabel_reward_fn=relabel_reward_fn,
            )

        if trajectory_path is None:
            # Setting the logger here is not necessary (PreferenceComparisons takes care
            # of it automatically) but it avoids creating unnecessary loggers.
            agent_trainer = preference_comparisons.AgentTrainer(
                algorithm=agent,
                reward_fn=reward_net,
                venv=venv,
                exploration_frac=exploration_frac,
                rng=_rnd,
                custom_logger=custom_logger,
                **trajectory_generator_kwargs,
            )
            # Stable Baselines will automatically occupy GPU 0 if it is available.
            # Let's use the same device as the SB3 agent for the reward model.
            reward_net = reward_net.to(agent_trainer.algorithm.device)
            trajectory_generator: preference_comparisons.TrajectoryGenerator = (
                agent_trainer
            )
        else:
            if exploration_frac > 0:
                raise ValueError(
                    "exploration_frac can't be set when a trajectory dataset is used",
                )
            trajectory_generator = preference_comparisons.TrajectoryDataset(
                trajectories=data_serialize.load_with_rewards(
                    trajectory_path,
                ),
                rng=_rnd,
                custom_logger=custom_logger,
                **trajectory_generator_kwargs,
            )

        fragmenter: preference_comparisons.Fragmenter = (
            preference_comparisons.RandomFragmenter(
                **fragmenter_kwargs,
                rng=_rnd,
                custom_logger=custom_logger,
            )
        )
        preference_model = preference_comparisons.PreferenceModel(
            **preference_model_kwargs,
            model=reward_net,
        )
        if active_selection:
            fragmenter = preference_comparisons.ActiveSelectionFragmenter(
                preference_model=preference_model,
                base_fragmenter=fragmenter,
                fragment_sample_factor=active_selection_oversampling,
                uncertainty_on=uncertainty_on,
                custom_logger=custom_logger,
            )
        gatherer = gatherer_cls(
            **gatherer_kwargs,
            rng=_rnd,
            custom_logger=custom_logger,
        )

        loss = preference_comparisons.CrossEntropyRewardLoss()

        reward_trainer = preference_comparisons._make_reward_trainer(
            preference_model,
            loss,
            _rnd,
            reward_trainer_kwargs,
        )

        main_trainer = preference_comparisons.PreferenceComparisons(
            trajectory_generator,
            reward_net,
            num_iterations=num_iterations,
            fragmenter=fragmenter,
            preference_gatherer=gatherer,
            reward_trainer=reward_trainer,
            comparison_queue_size=comparison_queue_size,
            fragment_length=fragment_length,
            transition_oversampling=transition_oversampling,
            initial_comparison_frac=initial_comparison_frac,
            custom_logger=custom_logger,
            allow_variable_horizon=allow_variable_horizon,
            query_schedule=query_schedule,
        )

        def save_callback(iteration_num):
            if checkpoint_interval > 0 and iteration_num % checkpoint_interval == 0:
                save_checkpoint(
                    trainer=main_trainer,
                    save_path=log_dir / "checkpoints" / f"{iteration_num:04d}",
                    allow_save_policy=bool(trajectory_path is None),
                )

        results = main_trainer.train(
            total_timesteps,
            total_comparisons,
            callback=save_callback,
        )

        # Storing and evaluating policy only useful if we generated trajectory data
        if bool(trajectory_path is None):
            results = dict(results)
            results["imit_stats"] = policy_evaluation.eval_policy(agent, venv)

    if save_preferences:
        main_trainer.dataset.save(log_dir / "preferences.pkl")

    # Save final artifacts.
    if checkpoint_interval >= 0:
        save_checkpoint(
            trainer=main_trainer,
            save_path=log_dir / "checkpoints" / "final",
            allow_save_policy=bool(trajectory_path is None),
        )

    return results


def main_console():
    observer_path = (
        pathlib.Path.cwd() / "output" / "sacred" / "train_preference_comparisons"
    )
    observer = FileStorageObserver(observer_path)
    train_preference_comparisons_ex.observers.append(observer)
    train_preference_comparisons_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



"""Implements the regularizer base class and some standard regularizers."""

import abc
from typing import Generic, Optional, Protocol, Type, TypeVar, Union

import numpy as np
import torch as th
from torch import optim

from imitation.regularization import updaters
from imitation.util import logger as imit_logger

# this is not actually a scalar, dimension check is still required for tensor.
Scalar = Union[th.Tensor, float]

R = TypeVar("R")
Self = TypeVar("Self", bound="Regularizer")
T_Regularizer_co = TypeVar(  # pytype: disable=not-supported-yet
    "T_Regularizer_co",
    covariant=True,
)


class RegularizerFactory(Protocol[T_Regularizer_co]):
    """Protocol for functions that create regularizers.

    The regularizer factory is meant to be used as a way to create a regularizer
    in two steps. First, the end-user creates a regularizer factory by calling
    the `.create()` method of a regularizer class. This allows specifying
    all the relevant configuration to the regularization algorithm. Then, the
    network algorithm finishes setting up the optimizer and logger, and calls
    the regularizer factory to create the regularizer.

    This two-step process separates the configuration of the regularization
    algorithm from additional "operational" parameters. This is useful because it
    solves two problems:

    #. The end-user does not have access to the optimizer and logger when
       configuring the regularization algorithm.
    #. Validation of the configuration is done outside the network constructor.

    It also allows re-using the same regularizer factory for multiple networks.
    """

    def __call__(
        self,
        *,
        optimizer: optim.Optimizer,
        logger: imit_logger.HierarchicalLogger,
    ) -> T_Regularizer_co:
        """Constructs a regularizer from the factory.

        Args:
            optimizer: The optimizer used by the network.
            logger: The logger used by the network.
        """


class Regularizer(abc.ABC, Generic[R]):
    """Abstract class for creating regularizers with a common interface."""

    optimizer: optim.Optimizer
    lambda_: float
    lambda_updater: Optional[updaters.LambdaUpdater]
    logger: imit_logger.HierarchicalLogger
    val_split: Optional[float]

    def __init__(
        self,
        optimizer: optim.Optimizer,
        initial_lambda: float,
        lambda_updater: Optional[updaters.LambdaUpdater],
        logger: imit_logger.HierarchicalLogger,
        val_split: Optional[float] = None,
    ) -> None:
        """Initialize the regularizer.

        Args:
            optimizer: The optimizer to which the regularizer is attached.
            initial_lambda: The initial value of the regularization parameter.
            lambda_updater: A callable object that takes in the current lambda and
                the train and val loss, and returns the new lambda.
            logger: The logger to which the regularizer will log its parameters.
            val_split: The fraction of the training data to use as validation data
                for the lambda updater. Can be none if no lambda updater is provided.

        Raises:
            ValueError: if no lambda updater (``lambda_updater``) is provided and the
                initial regularization strength (``initial_lambda``) is zero.
            ValueError: if a validation split (``val_split``) is provided but it's not a
                float in the (0, 1) interval.
            ValueError: if a lambda updater is provided but no validation split
                is provided.
            ValueError: if a validation split is set, but no lambda updater is
                provided.
        """
        if lambda_updater is None and np.allclose(initial_lambda, 0.0):
            raise ValueError(
                "If you do not pass a regularizer parameter updater your "
                "regularization strength must be non-zero, as this would "
                "result in no regularization.",
            )

        if val_split is not None and (
            not isinstance(val_split, float)
            or np.allclose(val_split, 0.0)
            or val_split <= 0
            or val_split >= 1
        ):
            raise ValueError(
                f"val_split = {val_split} must be a float strictly between 0 and 1.",
            )

        if lambda_updater is not None and val_split is None:
            raise ValueError(
                "If you pass a regularizer parameter updater, you must also "
                "specify a validation split. Otherwise the updater won't have any "
                "validation data to use for updating.",
            )
        elif lambda_updater is None and val_split is not None:
            raise ValueError(
                "If you pass a validation split, you must also "
                "pass a regularizer parameter updater. Otherwise you are wasting"
                " data into the validation split that will not be used.",
            )

        self.optimizer = optimizer
        self.lambda_ = initial_lambda
        self.lambda_updater = lambda_updater
        self.logger = logger
        self.val_split = val_split

        self.logger.record("regularization_lambda", self.lambda_)

    @classmethod
    def create(
        cls: Type[Self],
        initial_lambda: float,
        lambda_updater: Optional[updaters.LambdaUpdater] = None,
        val_split: float = 0.0,
        **kwargs,
    ) -> RegularizerFactory[Self]:
        """Create a regularizer."""

        def factory(
            *,
            optimizer: optim.Optimizer,
            logger: imit_logger.HierarchicalLogger,
        ) -> Self:
            return cls(
                initial_lambda=initial_lambda,
                optimizer=optimizer,
                lambda_updater=lambda_updater,
                logger=logger,
                val_split=val_split,
                **kwargs,
            )

        return factory

    @abc.abstractmethod
    def regularize_and_backward(self, loss: th.Tensor) -> R:
        """Abstract method for performing the regularization step.

        The return type is a generic and the specific implementation
        must describe the meaning of the return type.

        This step will also call `loss.backward()` for the user.
        This is because the regularizer may require the
        loss to be called before or after the regularization step.
        Leaving this to the user would force them to make their
        implementation dependent on the regularizer algorithm used,
        which is prone to errors.

        Args:
            loss: The loss to regularize.
        """

    def update_params(self, train_loss: Scalar, val_loss: Scalar) -> None:
        """Update the regularization parameter.

        This method calls the lambda_updater to update the regularization parameter,
        and assigns the new value to `self.lambda_`. Then logs the new value using
        the provided logger.

        Args:
            train_loss: The loss on the training set.
            val_loss: The loss on the validation set.
        """
        if self.lambda_updater is not None:
            self.lambda_ = self.lambda_updater(self.lambda_, train_loss, val_loss)
            self.logger.record("regularization_lambda", self.lambda_)


class LossRegularizer(Regularizer[Scalar]):
    """Abstract base class for regularizers that add a loss term to the loss function.

    Requires the user to implement the _loss_penalty method.
    """

    @abc.abstractmethod
    def _loss_penalty(self, loss: Scalar) -> Scalar:
        """Implement this method to add a loss term to the loss function.

        This method should return the term to be added to the loss function,
        not the regularized loss itself.

        Args:
            loss: The loss function to which the regularization term is added.
        """

    def regularize_and_backward(self, loss: th.Tensor) -> Scalar:
        """Add the regularization term to the loss and compute gradients.

        Args:
            loss: The loss to regularize.

        Returns:
            The regularized loss.
        """
        regularized_loss = th.add(loss, self._loss_penalty(loss))
        regularized_loss.backward()
        self.logger.record("regularized_loss", regularized_loss.item())
        return regularized_loss


class WeightRegularizer(Regularizer):
    """Abstract base class for regularizers that regularize the weights of a network.

    Requires the user to implement the _weight_penalty method.
    """

    @abc.abstractmethod
    def _weight_penalty(self, weight: th.Tensor, group: dict) -> Scalar:
        """Implement this method to regularize the weights of the network.

        This method should return the regularization term to be added to the weight,
        not the regularized weight itself.

        Args:
            weight: The weight (network parameter) to regularize.
            group: The group of parameters to which the weight belongs.
        """

    def regularize_and_backward(self, loss: th.Tensor) -> None:
        """Regularize the weights of the network, and call ``loss.backward()``."""
        loss.backward()
        for group in self.optimizer.param_groups:
            for param in group["params"]:
                param.data = th.add(param.data, self._weight_penalty(param, group))


class LpRegularizer(LossRegularizer):
    """Applies Lp regularization to a loss function."""

    p: int

    def __init__(
        self,
        optimizer: optim.Optimizer,
        initial_lambda: float,
        lambda_updater: Optional[updaters.LambdaUpdater],
        logger: imit_logger.HierarchicalLogger,
        p: int,
        val_split: Optional[float] = None,
    ) -> None:
        """Initialize the regularizer."""
        super().__init__(optimizer, initial_lambda, lambda_updater, logger, val_split)
        if not isinstance(p, int) or p < 1:
            raise ValueError("p must be a positive integer")
        self.p = p

    def _loss_penalty(self, loss: Scalar) -> Scalar:
        """Returns the loss penalty.

        Calculates the p-th power of the Lp norm of the weights in the optimizer,
        and returns a scaled version of it as the penalty.

        Args:
            loss: The loss to regularize.

        Returns:
            The scaled pth power of the Lp norm of the network weights.
        """
        del loss
        penalty = 0
        for group in self.optimizer.param_groups:
            for param in group["params"]:
                penalty += th.linalg.vector_norm(param, ord=self.p).pow(self.p)
        return self.lambda_ * penalty


class WeightDecayRegularizer(WeightRegularizer):
    """Applies weight decay to a loss function."""

    def _weight_penalty(self, weight, group) -> Scalar:
        """Returns the weight penalty.

        Args:
            weight: The weight to regularize.
            group: The group of parameters to which the weight belongs.

        Returns:
            The weight penalty (to add to the current value of the weight)
        """
        return -self.lambda_ * group["lr"] * weight.data



"""Logging for quantitative metrics and free-form text."""

import contextlib
import datetime
import os
import pathlib
import sys
import tempfile
from typing import Any, Dict, Generator, List, Optional, Sequence, Tuple

import stable_baselines3.common.logger as sb_logger

from imitation.data import types
from imitation.util import util


def make_output_format(
    _format: str,
    log_dir: str,
    log_suffix: str = "",
    max_length: int = 50,
) -> sb_logger.KVWriter:
    """Returns a logger for the requested format.

    Args:
        _format: the requested format to log to
            ('stdout', 'log', 'json' or 'csv' or 'tensorboard').
        log_dir: the logging directory.
        log_suffix: the suffix for the log file.
        max_length: the maximum length beyond which the keys get truncated.

    Returns:
        the logger.
    """
    os.makedirs(log_dir, exist_ok=True)
    if _format == "stdout":
        return sb_logger.HumanOutputFormat(sys.stdout, max_length=max_length)
    elif _format == "log":
        return sb_logger.HumanOutputFormat(
            os.path.join(log_dir, f"log{log_suffix}.txt"),
            max_length=max_length,
        )
    else:
        return sb_logger.make_output_format(_format, log_dir, log_suffix)


def _build_output_formats(
    folder: pathlib.Path,
    format_strs: Sequence[str],
) -> Sequence[sb_logger.KVWriter]:
    """Build output formats for initializing a Stable Baselines Logger.

    Args:
        folder: Path to directory that logs are written to.
        format_strs: A list of output format strings. For details on available
            output formats see `stable_baselines3.logger.make_output_format`.

    Returns:
        A list of output formats, one corresponding to each `format_strs`.
    """
    folder.mkdir(parents=True, exist_ok=True)
    output_formats: List[sb_logger.KVWriter] = []
    for f in format_strs:
        if f == "wandb":
            output_formats.append(WandbOutputFormat())
        else:
            output_formats.append(make_output_format(f, str(folder)))
    return output_formats


class HierarchicalLogger(sb_logger.Logger):
    """A logger supporting contexts for accumulating mean values.

    `self.accumulate_means` creates a context manager. While in this context,
    values are loggged to a sub-logger, with only mean values recorded in the
    top-level (root) logger.

    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as dir:
    ...     logger: HierarchicalLogger = configure(dir, ('log',))
    ...     # record the key value pair (loss, 1.0) to path `dir`
    ...     # at step 1.
    ...     logger.record("loss", 1.0)
    ...     logger.dump(step=1)
    ...     with logger.accumulate_means("dataset"):
    ...         # record the key value pair `("raw/dataset/entropy", 5.0)` to path
    ...         # `dir/raw/dataset` at step 100
    ...         logger.record("entropy", 5.0)
    ...         logger.dump(step=100)
    ...         # record the key value pair `("raw/dataset/entropy", 6.0)` to path
    ...         # `dir/raw/dataset` at step 200
    ...         logger.record("entropy", 6.0)
    ...         logger.dump(step=200)
    ...     # record the key value pair `("mean/dataset/entropy", 5.5)` to path
    ...     # `dir` at step 1.
    ...     logger.dump(step=1)
    ...     with logger.add_accumulate_prefix("foo"), logger.accumulate_means("bar"):
    ...         # record the key value pair ("raw/foo/bar/biz", 42.0) to path
    ...         # `dir/raw/foo/bar` at step 2000
    ...         logger.record("biz", 42.0)
    ...         logger.dump(step=2000)
    ...     # record the key value pair `("mean/foo/bar/biz", 42.0)` to path
    ...     # `dir` at step 1.
    ...     logger.dump(step=1)
    ...     with open(os.path.join(dir, 'log.txt')) as f:
    ...         print(f.read())
    -------------------
    | loss | 1        |
    -------------------
    ---------------------------------
    | mean/              |          |
    |    dataset/entropy | 5.5      |
    ---------------------------------
    -----------------------------
    | mean/          |          |
    |    foo/bar/biz | 42       |
    -----------------------------
    <BLANKLINE>
    """

    default_logger: sb_logger.Logger
    current_logger: Optional[sb_logger.Logger]
    _cached_loggers: Dict[str, sb_logger.Logger]
    _accumulate_prefixes: List[str]
    _key_prefixes: List[str]
    _subdir: Optional[str]
    _name: Optional[str]
    format_strs: Sequence[str]

    def __init__(
        self,
        default_logger: sb_logger.Logger,
        format_strs: Sequence[str] = ("stdout", "log", "csv"),
    ):
        """Builds HierarchicalLogger.

        Args:
            default_logger: The default logger when not in an `accumulate_means`
                context. Also the logger to which mean values are written to after
                exiting from a context.
            format_strs: A list of output format strings that should be used by
                every Logger initialized by this class during an `AccumulatingMeans`
                context. For details on available output formats see
                `stable_baselines3.logger.make_output_format`.
        """
        self.default_logger = default_logger
        self.current_logger = None
        self._cached_loggers = {}
        self._accumulate_prefixes = []
        self._key_prefixes = []
        self._subdir = None
        self._name = None
        self.format_strs = format_strs
        super().__init__(folder=self.default_logger.dir, output_formats=[])

    def _update_name_to_maps(self) -> None:
        self.name_to_value = self._logger.name_to_value
        self.name_to_count = self._logger.name_to_count
        self.name_to_excluded = self._logger.name_to_excluded

    @contextlib.contextmanager
    def add_accumulate_prefix(self, prefix: str) -> Generator[None, None, None]:
        """Add a prefix to the subdirectory used to accumulate means.

        This prefix only applies when a `accumulate_means` context is active. If there
        are multiple active prefixes, then they are concatenated.

        Args:
            prefix: The prefix to add to the named sub.

        Yields:
            None when the context manager is entered

        Raises:
            RuntimeError: if accumulate means context is already active.
        """
        if self.current_logger is not None:
            raise RuntimeError(
                "Cannot add prefix when accumulate_means context is already active.",
            )

        try:
            self._accumulate_prefixes.append(prefix)
            yield
        finally:
            self._accumulate_prefixes.pop()

    def get_accumulate_prefixes(self) -> str:
        prefixes = "/".join(self._accumulate_prefixes)
        return prefixes + "/" if prefixes else ""

    @contextlib.contextmanager
    def add_key_prefix(self, prefix: str) -> Generator[None, None, None]:
        """Add a prefix to the keys logged during an accumulate_means context.

        This prefix only applies when a `accumulate_means` context is active.
        If there are multiple active prefixes, then they are concatenated.

        Args:
            prefix: The prefix to add to the keys.

        Yields:
            None when the context manager is entered

        Raises:
            RuntimeError: if accumulate means context is already active.
        """
        if self.current_logger is None:
            raise RuntimeError(
                "Cannot add key prefix when accumulate_means context is not active.",
            )

        try:
            self._key_prefixes.append(prefix)
            yield
        finally:
            self._key_prefixes.pop()

    @contextlib.contextmanager
    def accumulate_means(self, name: str) -> Generator[None, None, None]:
        """Temporarily modifies this HierarchicalLogger to accumulate means values.

        Within this context manager, ``self.record(key, value)`` writes the "raw" values
        in ``f"{self.default_logger.log_dir}/[{accumulate_prefix}/]{name}"`` under the
        key ``"raw/[{accumulate_prefix}/]{name}/[{key_prefix}/]{key}"``, where
        ``accumulate_prefix`` is the concatenation of all prefixes added by
        ``add_accumulate_prefix`` and ``key_prefix`` is the concatenation of all
        prefixes added by ``add_key_prefix``, if any. At the same time, any call to
        ``self.record`` will also accumulate mean values on the default logger by
        calling::

            self.default_logger.record_mean(
                f"mean/[{accumulate_prefix}/]{name}/[{key_prefix}/]{key}",
                value,
            )

        Multiple prefixes may be active at once. In this case the `prefix` is simply the
        concatenation of each of the active prefixes in the order they
        were created e.g. if the active prefixes are ``['foo', 'bar']`` then
        the prefix is ``'foo/bar'``.

        After the context exits, calling ``self.dump()`` will write the means
        of all the "raw" values accumulated during this context to
        ``self.default_logger`` under keys of the form  ``mean/{prefix}/{name}/{key}``

        Note that the behavior of other logging methods, ``log`` and ``record_mean``
        are unmodified and will go straight to the default logger.

        Args:
            name: A string key which determines the ``folder`` where raw data is
                written and temporary logging prefixes for raw and mean data. Entering
                an `accumulate_means` context in the future with the same `subdir`
                will safely append to logs written in this folder rather than
                overwrite.

        Yields:
            None when the context is entered.

        Raises:
            RuntimeError: If this context is entered into while already in
                an `accumulate_means` context.
        """
        if self.current_logger is not None:
            raise RuntimeError("Nested `accumulate_means` context")

        subdir = os.path.join(*self._accumulate_prefixes, name)

        if subdir in self._cached_loggers:
            logger = self._cached_loggers[subdir]
        else:
            default_logger_dir = self.default_logger.dir
            assert default_logger_dir is not None
            folder = util.parse_path(default_logger_dir) / "raw" / subdir
            folder.mkdir(exist_ok=True, parents=True)
            output_formats = _build_output_formats(folder, self.format_strs)
            logger = sb_logger.Logger(str(folder), list(output_formats))
            self._cached_loggers[subdir] = logger
        try:
            self.current_logger = logger
            self._subdir = subdir
            self._name = name
            self._update_name_to_maps()
            yield
        finally:
            self.current_logger = None
            self._subdir = None
            self._name = None
            self._update_name_to_maps()

    def record(self, key, val, exclude=None):
        if self.current_logger is not None:  # In accumulate_means context.
            assert self._subdir is not None
            raw_key = "/".join(
                [
                    "raw",
                    *self._accumulate_prefixes,
                    self._name,
                    *self._key_prefixes,
                    key,
                ],
            )
            self.current_logger.record(raw_key, val, exclude)

            mean_key = "/".join(
                [
                    "mean",
                    *self._accumulate_prefixes,
                    self._name,
                    *self._key_prefixes,
                    key,
                ],
            )
            self.default_logger.record_mean(mean_key, val, exclude)
        else:  # Not in accumulate_means context.
            self.default_logger.record(key, val, exclude)

    @property
    def _logger(self):
        if self.current_logger is not None:
            return self.current_logger
        else:
            return self.default_logger

    def dump(self, step=0):
        self._logger.dump(step)

    def get_dir(self) -> str:
        return self._logger.get_dir()

    def log(self, *args, **kwargs):
        self.default_logger.log(*args, **kwargs)

    def set_level(self, level: int) -> None:
        self.default_logger.set_level(level)

    def record_mean(self, key, val, exclude=None):
        self.default_logger.record_mean(key, val, exclude)

    def close(self):
        self.default_logger.close()
        for logger in self._cached_loggers.values():
            logger.close()


class WandbOutputFormat(sb_logger.KVWriter):
    """A stable-baseline logger that writes to wandb.

    Users need to call `wandb.init()` before initializing `WandbOutputFormat`.
    """

    def __init__(self):
        """Initializes an instance of WandbOutputFormat.

        Raises:
            ModuleNotFoundError: wandb is not installed.
        """
        try:
            import wandb
        except ModuleNotFoundError as e:  # pragma: no cover
            raise ModuleNotFoundError(
                "Trying to log data with `WandbOutputFormat` "
                "but `wandb` not installed: try `pip install wandb`.",
            ) from e
        self.wandb_module = wandb

    def write(
        self,
        key_values: Dict[str, Any],
        key_excluded: Dict[str, Tuple[str, ...]],
        step: int = 0,
    ) -> None:
        for (key, value), (key_ex, excluded) in zip(
            sorted(key_values.items()),
            sorted(key_excluded.items()),
        ):
            assert key == key_ex, f"key mismatch between {key} and {key_ex}."
            if excluded is not None and "wandb" in excluded:
                continue

            self.wandb_module.log({key: value}, step=step)
        self.wandb_module.log({}, commit=True)

    def close(self) -> None:
        self.wandb_module.finish()


def configure(
    folder: Optional[types.AnyPath] = None,
    format_strs: Optional[Sequence[str]] = None,
) -> HierarchicalLogger:
    """Configure Stable Baselines logger to be `accumulate_means()`-compatible.

    After this function is called, `stable_baselines3.logger.{configure,reset}()`
    are replaced with stubs that raise RuntimeError.

    Args:
        folder: Argument from `stable_baselines3.logger.configure`.
        format_strs: An list of output format strings. For details on available
            output formats see `stable_baselines3.logger.make_output_format`.

    Returns:
        The configured HierarchicalLogger instance.
    """
    if folder is None:
        tempdir = util.parse_path(tempfile.gettempdir())
        now = datetime.datetime.now()
        timestamp = now.strftime("imitation-%Y-%m-%d-%H-%M-%S-%f")
        folder = tempdir / timestamp
    else:
        folder = util.parse_path(folder)
    if format_strs is None:
        format_strs = ["stdout", "log", "csv"]
    output_formats = _build_output_formats(folder, format_strs)
    default_logger = sb_logger.Logger(str(folder), list(output_formats))
    hier_format_strs = [f for f in format_strs if f != "wandb"]
    hier_logger = HierarchicalLogger(default_logger, hier_format_strs)
    return hier_logger



"""Tests interactive policies."""

import collections
from typing import cast
from unittest import mock

import gymnasium as gym
import numpy as np
import pytest
from gymnasium import spaces
from stable_baselines3.common import vec_env

from imitation.policies import interactive

SIMPLE_ENVS = [
    "seals/CartPole-v0",
]
ATARI_ENVS = [
    "Pong-v4",
]


class NoRenderingDiscreteInteractivePolicy(interactive.DiscreteInteractivePolicy):
    """DiscreteInteractivePolicy with no rendering."""

    def _render(self, obs: np.ndarray) -> None:
        pass


def _get_simple_interactive_policy(env: vec_env.VecEnv):
    num_actions = cast(spaces.Discrete, env.action_space).n
    action_keys_names = collections.OrderedDict(
        [(f"k{i}", f"n{i}") for i in range(num_actions)],
    )
    interactive_policy = NoRenderingDiscreteInteractivePolicy(
        env.observation_space,
        env.action_space,
        action_keys_names,
    )
    return interactive_policy


@pytest.mark.parametrize("env_name", SIMPLE_ENVS + ATARI_ENVS)
def test_interactive_policy(env_name: str):
    """Test if correct actions are selected, as specified by input keys."""
    env = vec_env.DummyVecEnv([lambda: gym.wrappers.TimeLimit(gym.make(env_name), 50)])
    env.seed(0)

    if env_name in ATARI_ENVS:
        interactive_policy = interactive.AtariInteractivePolicy(env)
    else:
        interactive_policy = _get_simple_interactive_policy(env)
    action_keys = list(interactive_policy.action_keys_names.keys())

    obs = env.reset()
    done = np.array([False])

    class mock_input:
        def __init__(self):
            self.index = 0

        def __call__(self, _):
            # Sometimes insert incorrect keys, which should get ignored by the policy.
            if np.random.uniform() < 0.5:
                return "invalid"
            key = action_keys[self.index]
            self.index = (self.index + 1) % len(action_keys)
            return key

    with mock.patch("builtins.input", mock_input()):
        requested_action = 0
        while not done.all():
            action, _ = interactive_policy.predict(np.array(obs))
            assert isinstance(action, np.ndarray)
            assert all(env.action_space.contains(a) for a in action)
            assert action[0] == requested_action

            obs, reward, done, info = env.step(action)
            assert isinstance(obs, np.ndarray)
            assert all(env.observation_space.contains(o) for o in obs)
            assert isinstance(reward, np.ndarray)
            assert isinstance(done, np.ndarray)

            requested_action = (requested_action + 1) % len(action_keys)


@pytest.mark.parametrize("env_name", SIMPLE_ENVS)
def test_interactive_policy_input_validity(capsys, env_name: str):
    """Test if appropriate feedback is given on the validity of the input."""
    env = vec_env.DummyVecEnv([lambda: gym.wrappers.TimeLimit(gym.make(env_name), 10)])
    env.seed(0)

    interactive_policy = _get_simple_interactive_policy(env)
    action_keys = list(interactive_policy.action_keys_names.keys())

    # Valid input key case
    obs = env.reset()

    def mock_input_valid(prompt):
        print(prompt)
        return action_keys[0]

    with mock.patch("builtins.input", mock_input_valid):
        interactive_policy.predict(obs)
        stdout = capsys.readouterr().out
        assert "Your choice" in stdout and "Invalid" not in stdout

    # First invalid input key, then valid
    obs = env.reset()

    class mock_input_invalid_then_valid:
        def __init__(self):
            self.return_valid = False

        def __call__(self, prompt):
            print(prompt)
            if self.return_valid:
                return action_keys[0]
            self.return_valid = True
            return "invalid"

    with mock.patch("builtins.input", mock_input_invalid_then_valid()):
        interactive_policy.predict(obs)
        stdout = capsys.readouterr().out
        assert "Your choice" in stdout and "Invalid" in stdout


@pytest.mark.parametrize("env_name", ATARI_ENVS)
def test_atari_action_mappings(env_name: str):
    """Test if correct actions are selected, as specified by input keys."""
    env = vec_env.DummyVecEnv([lambda: gym.wrappers.TimeLimit(gym.make(env_name), 50)])
    env.seed(0)
    action_meanings = env.env_method("get_action_meanings", indices=[0])[0]

    interactive_policy = interactive.AtariInteractivePolicy(env)

    obs = env.reset()

    provided_keys = ["2", "a", "d"]
    expected_action_meanings = ["FIRE", "LEFT", "RIGHT"]

    class mock_input:
        def __init__(self):
            self.index = 0

        def __call__(self, _):
            key = provided_keys[self.index]
            self.index += 1
            return key

    with mock.patch("builtins.input", mock_input()):
        for expected_action_meaning in expected_action_meanings:
            action, _ = interactive_policy.predict(np.array(obs))
            obs, reward, done, info = env.step(action)

            assert action_meanings[action[0]] == expected_action_meaning



#!/usr/bin/python3

"""This script starts an X server and sets DISPLAY, then runs wrapped command."""

# Usage: ./Xdummy-entrypoint.py [command]
#
# Adapted from https://github.com/openai/mujoco-py/blob/master/vendor/Xdummy-entrypoint
# Copyright OpenAI; MIT License

import argparse
import os
import subprocess
import sys

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    args, extra_args = parser.parse_known_args()

    subprocess.Popen(
        [
            "nohup",
            "Xorg",
            "-noreset",
            "+extension",
            "GLX",
            "+extension",
            "RANDR",
            "+extension",
            "RENDER",
            "-logfile",
            "/tmp/xdummy.log",
            "-config",
            "/etc/dummy_xorg.conf",
            ":0",
        ],
    )
    os.environ["DISPLAY"] = ":0"

    if not extra_args:
        argv = ["/bin/bash"]
    else:
        argv = extra_args

    # Explicitly flush right before the exec since otherwise things might get
    # lost in Python's buffers around stdout/stderr (!).
    sys.stdout.flush()
    sys.stderr.flush()

    os.execvpe(argv[0], argv, os.environ)



"""Custom policy classes and convenience methods."""

import abc
from typing import Dict, Type, Union

import gymnasium as gym
import numpy as np
import torch as th
from stable_baselines3.common import policies, torch_layers
from stable_baselines3.sac import policies as sac_policies
from torch import nn

from imitation.data import types
from imitation.util import networks


class NonTrainablePolicy(policies.BasePolicy, abc.ABC):
    """Abstract class for non-trainable (e.g. hard-coded or interactive) policies."""

    def __init__(self, observation_space: gym.Space, action_space: gym.Space):
        """Builds NonTrainablePolicy with specified observation and action space."""
        super().__init__(
            observation_space=observation_space,
            action_space=action_space,
        )

    def _predict(
        self,
        obs: Union[th.Tensor, Dict[str, th.Tensor]],
        deterministic: bool = False,
    ):
        np_actions = []
        if isinstance(obs, dict):
            np_obs: Union[types.DictObs, np.ndarray] = types.DictObs(
                {k: v.detach().cpu().numpy() for k, v in obs.items()},
            )
        else:
            np_obs = obs.detach().cpu().numpy()
        for np_ob in np_obs:
            np_ob_unwrapped = types.maybe_unwrap_dictobs(np_ob)
            assert self.observation_space.contains(np_ob_unwrapped)
            np_actions.append(self._choose_action(np_ob_unwrapped))
        np_actions = np.stack(np_actions, axis=0)
        th_actions = th.as_tensor(np_actions, device=self.device)
        return th_actions

    @abc.abstractmethod
    def _choose_action(
        self,
        obs: Union[np.ndarray, Dict[str, np.ndarray]],
    ) -> np.ndarray:
        """Chooses an action, optionally based on observation obs."""

    def forward(self, *args):
        # technically BasePolicy is a Torch module, so this needs a forward()
        # method
        raise NotImplementedError  # pragma: no cover


class RandomPolicy(NonTrainablePolicy):
    """Returns random actions."""

    def _choose_action(
        self,
        obs: Union[np.ndarray, Dict[str, np.ndarray]],
    ) -> np.ndarray:
        return self.action_space.sample()


class ZeroPolicy(NonTrainablePolicy):
    """Returns constant zero action."""

    def __init__(self, observation_space: gym.Space, action_space: gym.Space):
        """Builds ZeroPolicy with specified observation and action space."""
        super().__init__(observation_space, action_space)
        self._zero_action = np.zeros_like(
            action_space.sample(),
            dtype=action_space.dtype,
        )
        if self._zero_action not in action_space:
            raise ValueError(
                f"Zero action {self._zero_action} not in action space {action_space}",
            )

    def _choose_action(
        self,
        obs: Union[np.ndarray, Dict[str, np.ndarray]],
    ) -> np.ndarray:
        return self._zero_action


class FeedForward32Policy(policies.ActorCriticPolicy):
    """A feed forward policy network with two hidden layers of 32 units.

    This matches the IRL policies in the original AIRL paper.

    Note: This differs from stable_baselines3 ActorCriticPolicy in two ways: by
    having 32 rather than 64 units, and by having policy and value networks
    share weights except at the final layer, where there are different linear heads.
    """

    def __init__(self, *args, **kwargs):
        """Builds FeedForward32Policy; arguments passed to `ActorCriticPolicy`."""
        super().__init__(*args, **kwargs, net_arch=[32, 32])


class SAC1024Policy(sac_policies.SACPolicy):
    """Actor and value networks with two hidden layers of 1024 units respectively.

    This matches the implementation of SAC policies in the PEBBLE paper. See:
    https://arxiv.org/pdf/2106.05091.pdf
    https://github.com/denisyarats/pytorch_sac/blob/master/config/agent/sac.yaml

    Note: This differs from stable_baselines3 SACPolicy by having 1024 hidden units
    in each layer instead of the default value of 256.
    """

    def __init__(self, *args, **kwargs):
        """Builds SAC1024Policy; arguments passed to `SACPolicy`."""
        super().__init__(*args, **kwargs, net_arch=[1024, 1024])


class NormalizeFeaturesExtractor(torch_layers.FlattenExtractor):
    """Feature extractor that flattens then normalizes input."""

    def __init__(
        self,
        observation_space: gym.Space,
        normalize_class: Type[nn.Module] = networks.RunningNorm,
    ):
        """Builds NormalizeFeaturesExtractor.

        Args:
            observation_space: The space observations lie in.
            normalize_class: The class to use to normalize observations (after being
                flattened). This can be any Module that preserves the shape;
                e.g. `nn.BatchNorm*` or `nn.LayerNorm`.
        """
        super().__init__(observation_space)
        # Below we have to ignore the type error when initializing the class because
        # there is no simple way of specifying a protocol that admits one positional
        # argument for the number of features while being compatible with nn.Module.
        # (it would require defining a base class and forcing all the subclasses
        # to inherit from it).
        self.normalize = normalize_class(self.features_dim)  # type: ignore[call-arg]

    def forward(self, observations: th.Tensor) -> th.Tensor:
        flattened = super().forward(observations)
        return self.normalize(flattened)



"""Test utilities to conveniently generate expert trajectories."""
import math
import pathlib
import pickle
import shutil
import warnings
from os import PathLike
from pathlib import Path
from typing import Sequence

import huggingface_sb3 as hfsb3
import numpy as np
from filelock import FileLock
from torch.utils import data as th_data

import imitation.data.serialize as data_serialize
import imitation.policies.serialize as policies_serialize
from imitation.data import rollout, types, wrappers
from imitation.util import util


def generate_expert_trajectories(
    env_id: str,
    num_trajectories: int,
    rng: np.random.Generator,
) -> Sequence[types.TrajectoryWithRew]:  # pragma: no cover
    """Generate expert trajectories for the given environment.

    Note: will just pull a pretrained policy from the Hugging Face model hub.

    Args:
        env_id: The environment to generate trajectories for.
        num_trajectories: The number of trajectories to generate.
        rng: The random number generator to use.

    Returns:
        A list of trajectories with rewards.
    """
    env = util.make_vec_env(
        env_id,
        post_wrappers=[lambda e, _: wrappers.RolloutInfoWrapper(e)],
        rng=rng,
    )
    try:
        expert = policies_serialize.load_policy("ppo-huggingface", env, env_name=env_id)
        return rollout.rollout(
            expert,
            env,
            rollout.make_sample_until(min_episodes=num_trajectories),
            rng=rng,
        )
    finally:
        env.close()


def lazy_generate_expert_trajectories(
    cache_path: PathLike,
    env_id: str,
    num_trajectories: int,
    rng: np.random.Generator,
) -> Sequence[types.TrajectoryWithRew]:
    """Generate or load expert trajectories from cache.

    Args:
        cache_path: A path to the folder to be used as cache for the expert
            trajectories.
        env_id: The environment to generate trajectories for.
        num_trajectories: The number of trajectories to generate.
        rng: The random number generator to use.

    Returns:
        A list of trajectories with rewards.
    """
    environment_cache_path = pathlib.Path(cache_path) / hfsb3.EnvironmentName(env_id)
    environment_cache_path.mkdir(parents=True, exist_ok=True)

    trajectories_path = environment_cache_path / "rollout"

    # Note: we cast to str here because FileLock doesn't support pathlib.Path.
    with FileLock(str(environment_cache_path / "rollout.lock")):
        try:
            trajectories = data_serialize.load_with_rewards(trajectories_path)
        except (FileNotFoundError, pickle.PickleError) as e:  # pragma: no cover
            generation_reason = (
                "the cache is cold"
                if isinstance(e, FileNotFoundError)
                else "trajectory file format in the cache is outdated"
            )
            warnings.warn(
                f"Generating expert trajectories for {env_id} because "
                f"{generation_reason}.",
            )
            trajectories = generate_expert_trajectories(env_id, num_trajectories, rng)
            data_serialize.save(trajectories_path, trajectories)

    if len(trajectories) >= num_trajectories:
        return trajectories[:num_trajectories]
    else:  # pragma: no cover
        # If it is not enough, just throw away the cache and generate more.
        if trajectories_path.is_dir():
            # rmtree won't remove directory on Windows
            # until the last handle to the directory is closed
            del trajectories
            shutil.rmtree(trajectories_path)
        else:
            trajectories_path.unlink()
        return lazy_generate_expert_trajectories(
            cache_path,
            env_id,
            num_trajectories,
            rng,
        )


def make_expert_transition_loader(
    cache_dir: Path,
    batch_size: int,
    expert_data_type: str,
    env_name: str,
    rng: np.random.Generator,
    num_trajectories: int = 1,
    shuffle: bool = True,
):
    """Creates different kinds of PyTorch data loaders for expert transitions.

    Args:
        cache_dir: The directory to use for caching the expert trajectories.
        batch_size: The batch size to use for the data loader.
        expert_data_type: The type of expert data to use. Can be one of "trajectories",
            "data_loader", "ducktyped_data_loader", "transitions".
        env_name: The environment to generate trajectories for.
        rng: The random number generator to use.
        num_trajectories: The number of trajectories to generate.
        shuffle: Whether to shuffle the dataset when creating a data loader.

    Raises:
        ValueError: If `expert_data_type` is not one of the supported types.

    Returns:
        A pytorch data loader for expert transitions.
    """
    trajectories = lazy_generate_expert_trajectories(
        cache_dir,
        env_name,
        num_trajectories,
        rng,
    )

    if expert_data_type == "trajectories":
        return trajectories

    transitions = rollout.flatten_trajectories(trajectories)

    if len(transitions) < batch_size:  # pragma: no cover
        # If we have less transitions than the batch size, we estimate the trajectory
        # length and generate enough trajectories to fill the batch size.
        trajectory_length = len(transitions) // len(trajectories)
        min_required_trajectories = math.ceil(batch_size / trajectory_length)
        transitions = rollout.flatten_trajectories(
            lazy_generate_expert_trajectories(
                cache_dir,
                env_name,
                min_required_trajectories,
                rng,
            ),
        )

    if expert_data_type == "data_loader":
        return th_data.DataLoader(
            transitions,
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=True,
            collate_fn=types.transitions_collate_fn,
        )
    elif expert_data_type == "ducktyped_data_loader":

        class DucktypedDataset:
            """Used to check that any iterator over Dict[str, Tensor] works with BC."""

            def __init__(self, transitions: types.TransitionsMinimal, batch_size: int):
                """Builds `DucktypedDataset`."""
                self.trans = transitions
                self.batch_size = batch_size

            def __iter__(self):
                for start in range(
                    0,
                    len(self.trans) - self.batch_size,
                    self.batch_size,
                ):
                    end = start + self.batch_size
                    d = dict(
                        obs=self.trans.obs[start:end],
                        acts=self.trans.acts[start:end],
                    )
                    d = {k: util.safe_to_tensor(v) for k, v in d.items()}
                    yield d

        return DucktypedDataset(transitions, batch_size)
    elif expert_data_type == "transitions":
        return transitions
    else:  # pragma: no cover
        raise ValueError(f"Unexpected data type '{expert_data_type}'")



"""Wrapper to turn a policy into a more exploratory version."""

from typing import Dict, Optional, Tuple, Union

import numpy as np
from stable_baselines3.common import vec_env

from imitation.data import rollout
from imitation.util import util


class ExplorationWrapper:
    """Wraps a PolicyCallable to create a partially randomized version.

    This wrapper randomly switches between two policies: the wrapped policy,
    and a random one. After each action, the current policy is kept
    with a certain probability. Otherwise, one of these two policies is chosen
    at random (without any dependence on what the current policy is).

    The random policy uses the `action_space.sample()` method.
    """

    def __init__(
        self,
        policy: rollout.AnyPolicy,
        venv: vec_env.VecEnv,
        random_prob: float,
        switch_prob: float,
        rng: np.random.Generator,
        deterministic_policy: bool = False,
    ):
        """Initializes the ExplorationWrapper.

        Args:
            policy: The policy to randomize.
            venv: The environment to use (needed for sampling random actions).
            random_prob: The probability of picking the random policy when switching.
            switch_prob: The probability of switching away from the current policy.
            rng: The random state to use for seeding the environment and for
                switching policies.
            deterministic_policy: Whether to make the policy deterministic when not
                exploring. This must be False when ``policy`` is a ``PolicyCallable``.
        """
        policy_callable = rollout.policy_to_callable(policy, venv, deterministic_policy)
        self.wrapped_policy = policy_callable
        self.random_prob = random_prob
        self.switch_prob = switch_prob
        self.venv = venv

        self.rng = rng
        seed = util.make_seeds(self.rng)
        self.venv.action_space.seed(seed)

        self.current_policy = policy_callable
        # Choose the initial policy at random
        self._switch()

    def _random_policy(
        self,
        obs: Union[np.ndarray, Dict[str, np.ndarray]],
        state: Optional[Tuple[np.ndarray, ...]],
        episode_start: Optional[np.ndarray],
    ) -> Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]]:
        del state, episode_start  # Unused
        acts = [self.venv.action_space.sample() for _ in range(len(obs))]
        return np.stack(acts, axis=0), None

    def _switch(self) -> None:
        """Pick a new policy at random."""
        if self.rng.random() < self.random_prob:
            self.current_policy = self._random_policy
        else:
            self.current_policy = self.wrapped_policy

    def __call__(
        self,
        observation: Union[np.ndarray, Dict[str, np.ndarray]],
        input_state: Optional[Tuple[np.ndarray, ...]],
        episode_start: Optional[np.ndarray],
    ) -> Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]]:
        del episode_start  # Unused

        if input_state is not None:
            # This checks that we aren't passed a state
            raise ValueError("Exploration wrapper does not support stateful policies.")

        acts, output_state = self.current_policy(observation, None, None)

        if output_state is not None:
            # This checks that the policy doesn't return a state
            raise ValueError("Exploration wrapper does not support stateful policies.")

        if self.rng.random() < self.switch_prob:
            self._switch()
        return acts, None



"""Train GAIL or AIRL."""

import functools
import logging
import pathlib
from typing import Any, Mapping, Optional, Type

import sacred.commands
import torch as th
from sacred.observers import FileStorageObserver

from imitation.algorithms.adversarial import airl as airl_algo
from imitation.algorithms.adversarial import common
from imitation.algorithms.adversarial import gail as gail_algo
from imitation.data import rollout
from imitation.policies import serialize
from imitation.scripts.config.train_adversarial import train_adversarial_ex
from imitation.scripts.ingredients import demonstrations, environment
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients import policy_evaluation, reward, rl

logger = logging.getLogger("imitation.scripts.train_adversarial")


def save(trainer: common.AdversarialTrainer, save_path: pathlib.Path):
    """Save discriminator and generator."""
    # We implement this here and not in Trainer since we do not want to actually
    # serialize the whole Trainer (including e.g. expert demonstrations).
    save_path.mkdir(parents=True, exist_ok=True)
    th.save(trainer.reward_train, save_path / "reward_train.pt")
    th.save(trainer.reward_test, save_path / "reward_test.pt")
    serialize.save_stable_model(
        save_path / "gen_policy",
        trainer.gen_algo,
    )


def _add_hook(ingredient: sacred.Ingredient) -> None:
    # This is an ugly hack around Sacred config brokenness.
    # Config hooks only apply to their current ingredient,
    # and cannot update things in nested ingredients.
    # So we have to apply this hook to every ingredient we use.
    @ingredient.config_hook
    def hook(config, command_name, logger):
        del logger
        path = ingredient.path
        if path == "train_adversarial":
            path = ""
        ingredient_config = sacred.utils.get_by_dotted_path(config, path)
        return ingredient_config["algorithm_specific"].get(command_name, {})

    # We add this so Sacred doesn't complain that algorithm_specific is unused
    @ingredient.capture
    def dummy_no_op(algorithm_specific):
        pass

    # But Sacred may then complain it isn't defined in config! So, define it.
    @ingredient.config
    def dummy_config():
        algorithm_specific = {}  # noqa: F841


for ingredient in [train_adversarial_ex, *train_adversarial_ex.ingredients]:
    _add_hook(ingredient)


@train_adversarial_ex.capture
def train_adversarial(
    _run,
    show_config: bool,
    algo_cls: Type[common.AdversarialTrainer],
    algorithm_kwargs: Mapping[str, Any],
    total_timesteps: int,
    checkpoint_interval: int,
    agent_path: Optional[str],
) -> Mapping[str, Mapping[str, float]]:
    """Train an adversarial-network-based imitation learning algorithm.

    Checkpoints:
        - AdversarialTrainer train and test RewardNets are saved to
           `f"{log_dir}/checkpoints/{step}/reward_{train,test}.pt"`
            where step is either the training round or "final".
        - Generator policies are saved to `f"{log_dir}/checkpoints/{step}/gen_policy/"`.

    Args:
        show_config: Print the merged config before starting training. This is
            analogous to the print_config command, but will show config after
            rather than before merging `algorithm_specific` arguments.
        algo_cls: The adversarial imitation learning algorithm to use.
        algorithm_kwargs: Keyword arguments for the `GAIL` or `AIRL` constructor.
        total_timesteps: The number of transitions to sample from the environment
            during training.
        checkpoint_interval: Save the discriminator and generator models every
            `checkpoint_interval` rounds and after training is complete. If 0,
            then only save weights after training is complete. If <0, then don't
            save weights at all.
        agent_path: Path to a directory containing a pre-trained agent. If
            provided, then the agent will be initialized using this stored policy
            (warm start). If not provided, then the agent will be initialized using
            a random policy.

    Returns:
        A dictionary with two keys. "imit_stats" gives the return value of
        `rollout_stats()` on rollouts test-reward-wrapped environment, using the final
        policy (remember that the ground-truth reward can be recovered from the
        "monitor_return" key). "expert_stats" gives the return value of
        `rollout_stats()` on the expert demonstrations.
    """
    # This allows to specify total_timesteps and checkpoint_interval in scientific
    # notation, which is interpreted as a float by python.
    total_timesteps = int(total_timesteps)
    checkpoint_interval = int(checkpoint_interval)

    if show_config:
        # Running `train_adversarial print_config` will show unmerged config.
        # So, support showing merged config from `train_adversarial {airl,gail}`.
        sacred.commands.print_config(_run)

    custom_logger, log_dir = logging_ingredient.setup_logging()
    expert_trajs = demonstrations.get_expert_trajectories()

    with environment.make_venv() as venv:  # type: ignore[wrong-arg-count]
        reward_net = reward.make_reward_net(venv)
        relabel_reward_fn = functools.partial(
            reward_net.predict_processed,
            update_stats=False,
        )

        if agent_path is None:
            gen_algo = rl.make_rl_algo(venv, relabel_reward_fn=relabel_reward_fn)
        else:
            gen_algo = rl.load_rl_algo_from_path(
                agent_path=agent_path,
                venv=venv,
                relabel_reward_fn=relabel_reward_fn,
            )

        logger.info(f"Using '{algo_cls}' algorithm")
        algorithm_kwargs = dict(algorithm_kwargs)
        for k in ("shared", "airl", "gail"):
            # Config hook has copied relevant subset of config to top-level.
            # But due to Sacred limitations, cannot delete the rest of it.
            # So do that here to avoid passing in invalid arguments to constructor.
            if k in algorithm_kwargs:
                del algorithm_kwargs[k]
        trainer = algo_cls(
            venv=venv,
            demonstrations=expert_trajs,
            gen_algo=gen_algo,
            log_dir=log_dir,
            reward_net=reward_net,
            custom_logger=custom_logger,
            **algorithm_kwargs,
        )

        def callback(round_num: int, /) -> None:
            if checkpoint_interval > 0 and round_num % checkpoint_interval == 0:
                save(trainer, log_dir / "checkpoints" / f"{round_num:05d}")

        trainer.train(total_timesteps, callback)
        imit_stats = policy_evaluation.eval_policy(trainer.policy, trainer.venv_train)

    # Save final artifacts.
    if checkpoint_interval >= 0:
        save(trainer, log_dir / "checkpoints" / "final")

    return {
        "imit_stats": imit_stats,
        "expert_stats": rollout.rollout_stats(expert_trajs),
    }


@train_adversarial_ex.command
def gail():
    return train_adversarial(algo_cls=gail_algo.GAIL)


@train_adversarial_ex.command
def airl():
    return train_adversarial(algo_cls=airl_algo.AIRL)


def main_console():
    observer_path = pathlib.Path.cwd() / "output" / "sacred" / "train_adversarial"
    observer = FileStorageObserver(observer_path)
    train_adversarial_ex.observers.append(observer)
    train_adversarial_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()
